'''
# Flybys in `REBOUND`
- A package for implementing flybys in [hannorein/rebound](https://github.com/hannorein/rebound)
'''
from .core import (add_star_to_sim, flyby, flybys, hybrid_flyby, hybrid_flybys, successive_flybys, concurrent_flybys)
from .environments import (StellarEnvironment, LocalNeighborhood, OpenCluster, GlobularCluster, GalacticBulge, GalacticCore)
from .analytic import (relative_energy_change, energy_change_adiabatic_estimate, eccentricity_change_adiabatic_estimate, inclination_change_adiabatic_estimate)
from .imf import (IMF)
from .stars import (Star, Stars)
from .units import (UnitSet)

__version__ = 'v0.7.0'



import numpy as _np
import joblib as _joblib
import warnings as _warnings
from scipy.special import j0 as _j0,jv as _jv

from . import tools as _tools
from . import units as _u
from .core import add_star_to_sim, _rotate_into_plane

############################################################
################## Energy Estimates ########################
############################################################

def binary_energy(sim, particle_index=1):
    """
    Calculate the energy of a binary system, $-(G M m)/(2 a)$.

    Args:
        sim (Simulation): The simulation with two bodies, a central star and a planet.
        particle_index (int, optional): The index of the particle in the simulation to calculate the energy for. Default is 1.

    Returns:
        energy (Quantity): The energy of the binary system.

    Example:
        ```python
        import rebound
        import airball
        sim = rebound.Simulation()
        sim.add(m=1.)
        sim.add(m=5e-5, a=30., e=0.1, inc=0.1)
        energy = airball.analytic.binary_energy(sim)
        ```
    """
    index = int(particle_index)
    unit_set = _tools.rebound_units(sim)
    G = (sim.G * unit_set.length**3 / unit_set.mass / unit_set.time**2)
    p = sim.particles
    if p[index].m == 0:
        message = "Planet has zero mass therefore the binary energy is zero. This may cause divide by zero error when calculating the relative change in energy."
        _warnings.warn(message, RuntimeWarning)
    return (-G * p[0].m * unit_set.mass * p[index].m * unit_set.mass / (2. * p[index].a * unit_set['length'])).decompose(list(unit_set.values()))
    
def energy_change_adiabatic_estimate(sim, star, averaged=False, particle_index=1):
    """
    An analytical estimate for the change in energy of a binary system due to a stellar flyby.

    This function is based on the conclusions of [Roy & Haddow (2003)](https://ui.adsabs.harvard.edu/abs/2003CeMDA..87..411R/abstract) and [Heggie (2006)](https://ui.adsabs.harvard.edu/abs/2006fbp..book...20H/abstract). The orbital element angles of the flyby star are determined with respect to the plane defined by the binary orbit.

    Args:
        sim (Simulation): The simulation to calculate the energy change for.
        star (Star or Stars): The star or stars that are flying by the binary system.
        averaged (bool, optional): Whether to return the averaged energy change. Default is False.
        particle_index (int, optional): The index of the particle in the simulation to calculate the energy change for. Default is 1.

    Returns:
        result (Quantity): The estimated change in energy.

    Example:
        ```python
        import rebound
        import airball
        sim = rebound.Simulation()
        sim.add(m=1.)
        sim.add(m=5e-5, a=30., e=0.1, inc=0.1)
        star = airball.Star(m=1., b=500, v=5)
        energy_change = airball.energy_change_adiabatic_estimate(sim, star)
        ```
    """
    index = int(particle_index)
    units = _tools.rebound_units(sim)
    t0 = 0*units.time
    G = (sim.G * units.length**3 / units.mass / units.time**2)

    sim = sim.copy()
    _rotate_into_plane(sim, plane=index)
    # add_star_to_sim(sim, star, hash='flybystar', rmax=0) # Initialize Star at perihelion
    sim.move_to_hel()
    
    p = sim.particles
    m1, m2, m3 = p[0].m * units.mass, p[index].m * units.mass, star.mass # redefine the masses for convenience
    M12 = m1 + m2 # total mass of the binary system
    M123 = m1 + m2 + m3 # total mass of all the objects involved
    
    mu = G * (_tools.system_mass(sim)  * units.mass + m3)
    es = _tools.calculate_eccentricity(sim, star)
    
    a, e = p[index].a * units.length, p[index].e # redefine the orbital elements of the planet for convenience
    # Case: Non-circular Binary

    n = _np.sqrt(G*M12/a**3) # compute the mean motion of the planet
    
    w, W, i = star.omega, star.Omega, star.inc # redefine the orientation elements of the flyby star for convenience
    V = star.v
    # GM123 = G*M123 
    q = (- mu + _np.sqrt( mu**2. + star.b**2. * V**4.))/V**2. # compute the periapsis of the flyby star

    # Calculate the following convenient functions of the planet's eccentricity and Bessel functions of the first kind of order n.
    e1 = _jv(-1,e) - 2*e*_j0(e) + 2*e*_jv(2,0) - _jv(3,e)
    e2 = _jv(-1,e) - _jv(3,e)
    e4 = _jv(-1,e) - e*_j0(e) - e*_jv(2,e) + _jv(3,e)

    # Calculate a convenient function of the planet's semi-major axis and the flyby star's periapsis.
    k = _np.sqrt((2*M12*q**3)/(M123*a**3))
    
    # Calculate convenient functions of the flyby star's eccentricity.
    f1 = ((es + 1.0)**(0.75)) / ((2.0**(0.75)) * (es*es))
    with _u.set_enabled_equivalencies(_u.dimensionless_angles()):
        f2 = (3.0/(2.0*_np.sqrt(2.0))) * (_np.sqrt((es*es) - 1.0) - _np.arccos(1.0/es)) / ((es - 1.0)**(1.5))
    
    # Compute the prefactor and terms of the calculation done by Roy & Haddow (2003)
    prefactor = (-_np.sqrt(_np.pi)/8.0) * ((G*m1*m2*m3)/M12) * ((a*a)/(q*q*q)) * f1 * k**(2.5) * _np.exp((-2.0*k/3.0)*f2)
    with _u.set_enabled_equivalencies(_u.dimensionless_angles()):
        term1 = e1 * ( _np.sin(2.0*w + n*t0)*_np.cos(2.0*i - 1.0)- _np.sin(2.0*w + n*t0)*_np.cos(2.0*i)*_np.cos(2.0*W) - 3.0*_np.sin(n*t0 + 2.0*w)*_np.cos(2.0*W) - 4.0*_np.sin(2.0*W)*_np.cos(2.0*w + n*t0)*_np.cos(i) )
        term2 = e2 * (1.0 - e*e) * ( _np.sin(2.0*w + n*t0)*(1.0-_np.cos(2.0*i)) - _np.sin(2.0*w + n*t0)*_np.cos(2.0*i)*_np.cos(2.0*W) - 3.0*_np.sin(n*t0 +2.0*w)*_np.cos(2.0*W) - 4.0*_np.cos(n*t0 + 2.0*w)*_np.sin(2.0*W)*_np.cos(i) )
        term3 = e4 * _np.sqrt(1.0 - e*e) * (-2.0*_np.cos(2.0*i)*_np.cos(2.0*w + n*t0)*_np.sin(2.0*W) - 6.0*_np.cos(2.0*w + n*t0)*_np.sin(2.0*W) - 8.0*_np.cos(2.0*W)*_np.sin(2.0*w + n*t0)*_np.cos(i) )
    
    noncircular_result = None
    if averaged: noncircular_result = (prefactor * (e1 + e2 * (1 - e*e) + 2 * e4 * _np.sqrt(1 - e*e))).decompose(list(units.values()))
    else: noncircular_result = (prefactor * ( term1 + term2 + term3)).decompose(list(units.values()))

    # Case: Circular Binary

    # Compute the prefactor and terms of the calculation done by Roy & Haddow (2003)
    prefactor = (-_np.sqrt(_np.pi)/8.0) * ((G*m1*m2*m3)/M12) * ((a*a*a)/(q*q*q*q)) * f1 * k**(3.5) * _np.exp((-2.0*k/3.0)*f2) * (m2/M12 - m1/M12)
    with _u.set_enabled_equivalencies(_u.dimensionless_angles()):
        term1 = (1.0 + _np.cos(i)) * _np.sin(i)**2.0
        term2 = ( (_np.cos(w)**3.0) - 3.0 * (_np.sin(w)**2.0) * _np.cos(w) ) * _np.sin(n*t0)
        term3 = ( 3.0 * (_np.cos(w)**2.0) * _np.sin(w) - (_np.sin(w)**3.0)) * _np.cos(n*t0)
    
    circular_result = None
    if averaged: circular_result = (prefactor / _np.pi).decompose(list(units.values()))
    else: circular_result = (prefactor * term1 * (term2 + term3)).decompose(list(units.values()))

    return circular_result + noncircular_result

def energy_change_close_encounter_estimate(sim, star, particle_index=1):
    '''
    An analytical estimate for the change in energy of a binary system due to a flyby star. From Equation (4.7) of [Heggie (1975)](https://ui.adsabs.harvard.edu/abs/1975MNRAS.173..729H/abstract). The orbital element angles of the flyby star are determined with respect to the plane defined by the binary orbit.

    Args:
        sim (Simulation): The simulation with two bodies, a central star and a planet.
        star (Star): The star used to estimate the change in energy of a binary due to the flyby star
        particle_index (int, optional): The index of the particle in the simulation to calculate the energy for. Default is 1.

    Returns:
        result (Quantity): The change in energy of the binary system.

    Example:
        ```python
        import rebound
        from airball import Star
        sim = rebound.Simulation()
        sim.add(m=1.)
        sim.add(m=5e-5, a=30., e=0.1, inc=0.1)
        star = Star(m=1., b=100, v=5)
        energy_change = airball.analytic.energy_change_close_encounter_estimate(sim, star)
        ```
    '''
    index = int(particle_index)
    units = _tools.rebound_units(sim)
    
    def process_star(this_sim, this_star, this_index):
        s = this_sim.copy()
        add_star_to_sim(s, this_star, hash='flybystar', rmax=0, plane=this_index) # Initialize Star at perihelion
        return _tools.unit_vector(s.particles['flybystar'].xyz << units.length)
    
    if star.N == 1:
        x,y,z = process_star(sim, star, index)
    else:
        x,y,z = _np.asarray(_joblib.Parallel(n_jobs=-1)(_joblib.delayed(process_star)(sim, this_star, index) for this_star in star)).T
    c = _tools.cartesian_elements(sim, star, rmax=0)
    dat = _np.array([c['x'].value, c['y'].value, c['z'].value]).T << units.length
    G = (sim.G * units.length**3 / units.mass / units.time**2) # Newton's Gravitational constant

    m1, m2 = sim.particles[0].m * units.mass, sim.particles[index].m * units.mass # redefine the masses for convenience
    m3 = star.m
    M12 = m1 + m2 # total mass of the binary system
    M23 = m2 + m3 # total mass of the second and third bodies

    V = star.v # velocity of the star

    
    vx,vy,vz = sim.particles[index].vxyz << (units.length/units.time)

    with _u.set_enabled_equivalencies(_u.dimensionless_angles()):
        cosϕ = 1.0/_np.sqrt(1.0 + (((star.b**2.0)*(V**4.0))/((G*M23)**2.0)).decompose())
    
    prefactor = (-2.0 * m1 * m2 * m3)/(M12 * M23) * V * cosϕ
    t1 = -(x*vx + y*vy + z*vz)
    t2 = (m3 * V * cosϕ)/M23
    return (prefactor * (t1 + t2)).decompose(units.values())

def relative_energy_change_close_encounter_estimate(sim, star, particle_index=1):
    '''
     A convenience function for computing the analytical estimate for the relative change in energy of a binary system due to a close encounter with a flyby star. Combines [`energy_change_close_encounter_estimate`][airball.analytic.energy_change_close_encounter_estimate] and [`binary_energy`][airball.analytic.binary_energy] functions.

    Args:
        sim (Simulation): The simulation with two bodies, a central star and a planet.
        star (Star or Stars): The star or stars used to estimate the change in energy of a binary due to the flyby star
        particle_index (int, optional): The index of the particle in the simulation to calculate the energy for. Default is 1.

    Returns:
        energy_change (Quantity): The relative change in energy of the binary system.

    Example:
        ```python
        import rebound
        import airball
        sim = rebound.Simulation()
        sim.add(m=1.)
        sim.add(m=5e-5, a=30., e=0.1, inc=0.1)
        star = airball.Star(m=1., b=500, v=5)
        energy_change = airball.analytic.relative_energy_change_close_encounter_estimate(sim, star)
        ```

    '''
    return energy_change_close_encounter_estimate(sim, star, particle_index=particle_index)/binary_energy(sim, particle_index=particle_index)

def relative_energy_change(sim, star, averaged=False, particle_index=1):
    '''
     A convenience function for computing the analytical estimate for the relative change in energy of a binary system due to a flyby star. Combines [`energy_change_adiabatic_estimate`][airball.analytic.energy_change_adiabatic_estimate] and [`binary_energy`][airball.analytic.binary_energy] functions.

    Args:
        sim (Simulation): The simulation with two bodies, a central star and a planet.
        star (Star or Stars): The star or stars used to estimate the change in energy of a binary due to the flyby star
        averaged (bool, optional): Boolean indicator for averaging over the incident angles of the flyby star. Default is False.
        particle_index (int, optional): The index of the particle in the simulation to calculate the energy for. Default is 1.

    Returns:
        energy_change (Quantity): The relative change in energy of the binary system.

    Example:
        ```python
        import rebound
        import airball
        sim = rebound.Simulation()
        sim.add(m=1.)
        sim.add(m=5e-5, a=30., e=0.1, inc=0.1)
        star = airball.Star(m=1., b=500, v=5)
        energy_change = airball.analytic.relative_energy_change(sim, star)
        ```

    '''
    unit_set = _tools.rebound_units(sim)
    qs = star.q(sim)/(sim.particles[particle_index].a * unit_set.length)
    q_crit = 2**(2/3)
    if star.N == 1:
        if qs < q_crit: return energy_change_close_encounter_estimate(sim, star, particle_index=particle_index)/binary_energy(sim, particle_index=particle_index)
        else: return energy_change_adiabatic_estimate(sim, star, averaged=averaged, particle_index=particle_index)/binary_energy(sim, particle_index=particle_index)
    else:
        de_close = energy_change_close_encounter_estimate(sim, star, particle_index=particle_index)/binary_energy(sim, particle_index=particle_index)
        de = energy_change_adiabatic_estimate(sim, star, averaged=averaged, particle_index=particle_index)/binary_energy(sim, particle_index=particle_index)
        close_inds = qs < q_crit
        far_inds = ~close_inds
        results = _np.zeros(star.N)
        results[close_inds] = de_close[close_inds]
        results[far_inds] = de[far_inds]
        return results
    # return energy_change_adiabatic_estimate(sim=sim, star=star, averaged=averaged, particle_index=particle_index)/binary_energy(sim, particle_index=particle_index)

def parallel_relative_energy_change(sims, stars, averaged=False, particle_index=1):
    '''
        A convenience function for computing the analytical estimate for the relative change in energy of a binary system due to a flyby star. Combines [`energy_change_adiabatic_estimate`][airball.analytic.energy_change_adiabatic_estimate] and [`binary_energy`][airball.analytic.binary_energy] functions parallelized over the input parameters.

        Args:
            sims (list of Simulations): The simulation with two bodies, a central star and a planet.
            stars (Stars): The stars used to estimate the change in energy of a binary due to the flyby star
            averaged (bool, optional): Boolean indicator for averaging over the incident angles of the flyby star. Default is False.
            particle_index (int, optional): The index of the particle in the simulation to calculate the energy for. Default is 1.

        Returns:
            result (list of Quantities): The relative change in energy of the binary systems.

        Example:
            ```python
            import rebound
            import airball
            def setup():
                sim = rebound.Simulation()
                sim.add(m=1.)
                sim.add(m=5e-5, a=30., e=0.1, f='uniform')
                return sim
            stars = airball.Star(m=1., b=500, v=5, size=10)
            sims = [setup() for _ in range(stars.N)]
            energy_change = airball.analytic.parallel_relative_energy_change(sims, stars)
            ```
    '''
    unit_set = _tools.rebound_units(sims[0])
    qs = stars.q(sims[0])
    de = _np.array(_joblib.Parallel(n_jobs=-1)(_joblib.delayed(relative_energy_change)(sim=sims[i], star=stars[i], averaged=averaged, particle_index=particle_index) for i in range(stars.N)))
    de_close = _np.array(_joblib.Parallel(n_jobs=-1)(_joblib.delayed(relative_energy_change_close_encounter_estimate)(sim=sims[i], star=stars[i], particle_index=particle_index) for i in range(stars.N)))
    close_inds = qs/(sims[0].particles[particle_index].a * unit_set.length) < 2**(2/3)
    far_inds = ~close_inds
    results = _np.zeros(stars.N)
    results[close_inds] = de_close[close_inds]
    results[far_inds] = de[far_inds]
    return results


############################################################
################ Eccentricity Estimates ####################
############################################################


def eccentricity_change_close_encounter_estimate(sim, star, particle_index=1):
    '''
    An analytical estimate for the change in eccentricity of a binary system due to a flyby star. From Equation A1 of [Spurzem et al. (2009)](https://ui.adsabs.harvard.edu/abs/2009MNRAS.393..457S/abstract). The orbital element angles of the flyby star are determined with respect to the plane defined by the binary orbit.
    '''

    index = int(particle_index)
    unit_set = _tools.rebound_units(sim)

    p = sim.particles
    m1, m2, m3 = p[0].m * unit_set.mass, p[index].m * unit_set.mass, star.mass # redefine the masses for convenience
    M12 = m1 + m2 # total mass of the binary system
    M123 = m1 + m2 + m3 # total mass of all the objects involved

    q = star.q(sim) # compute the periapsis of the flyby star
    AB = _tools.hyperbolic_plane(sim, star)
    A,B = AB['A'], AB['B']
    ahat, bhat = [1,0,0], [0,1,0]

    # mu = G * (_tools.system_mass(sim)  * unit_set.mass + m3)
    # es = _tools.vinf_and_b_to_e(mu=mu, star_b=star.b, star_v=star.v)

    a, e = p[index].a * unit_set.length, p[index].e # redefine the orbital elements of the planet for convenience

    return ( ((-15.0*_np.pi*m3) / _np.sqrt(32.0 * M12*M123)) * ((a/q)**1.5) * (e * _np.sqrt(1.0 - e*e)) * ( (ahat @ A) * (bhat @ A) + (ahat @ B) * (bhat @ B) ) ).decompose(list(unit_set.values()))

def eccentricity_change_adiabatic_estimate(sim, star, averaged=False, particle_index=1, rmax=1e5*_u.au):
    '''
        An analytical estimate for the change in eccentricity of a binary system due to a flyby star. From Equations (7, 9, & 12) of Heggie & Rasio [(1996)](https://ui.adsabs.harvard.edu/abs/1996MNRAS.282.1064H/abstract). The orbital element angles of the flyby star are determined with respect to the plane defined by the binary orbit.

        Args:
            sim (Simulation): The simulation with two bodies, a central star and a planet.
            star (Star): The star used to estimate the change in energy of a binary due to the flyby star
            averaged (bool, optional): Boolean indicator for averaging over the incident angles of the flyby star. Default is False.
            particle_index (int, optional): The index of the particle in the simulation to calculate the energy for. Default is 1.
            rmax (float or Quantity, optional): Needed to determine the time when the star will be at perihelion. Default is $10^5$ AU.

        Returns:
            result (Quantity): The change in eccentricity of the binary system.

        Example:
            ```python
            import rebound
            import airball
            sim = rebound.Simulation()
            sim.add(m=1.)
            sim.add(m=5e-5, a=30., e=0.1, inc=0.1)
            star = airball.Star(m=1., b=500, v=5)
            eccentricity_change = airball.analytic.eccentricity_change_adiabatic_estimate(sim, star)
            ```
    '''

    index = int(particle_index)

    unit_set = _tools.rebound_units(sim)

    t0 = sim.t * unit_set.time
    G = (sim.G * unit_set.length**3 / unit_set.mass / unit_set.time**2)
    
    p = sim.particles
    m1, m2, m3 = p[0].m * unit_set.mass, p[index].m * unit_set.mass, star.mass # redefine the masses for convenience
    M12 = m1 + m2 # total mass of the binary system
    M123 = m1 + m2 + m3 # total mass of all the objects involved
    
    mu = G * (_tools.system_mass(sim)  * unit_set.mass + m3)
    es = _tools.vinf_and_b_to_e(mu=mu, star_b=star.b, star_v=star.v)
    
    a, e = p[index].a * unit_set.length, p[index].e # redefine the orbital elements of the planet for convenience
    # n = _np.sqrt(G*M12/a**3) # compute the mean motion of the planet
    
    w, W, i = star.omega, star.Omega, star.inc # redefine the orientation elements of the flyby star for convenience

    star_params = _tools.hyperbolic_elements(sim, star, rmax)
    tperi = star_params['T']
    Mperi = p[index].M + (p[index].n/unit_set.time) * (tperi - t0) # get the Mean anomaly when the flyby star is at perihelion
    f = _tools.M_to_f(p[index].e, Mperi.value) << _u.rad # get the true anomaly when the flyby star is at perihelion
    Wp = W + f
    V = star.v
    q = (- mu + _np.sqrt( mu**2. + star.b**2. * V**4.))/V**2. # compute the periapsis of the flyby star

    # Case: Non-Circular Binary

    prefactor = (-15.0/4.0) * m3 / _np.sqrt(M12*M123) * ((a/q)**1.5) * ((e * _np.sqrt(1.0 - e*e))/((1.0 + es)**1.5))
    with _u.set_enabled_equivalencies(_u.dimensionless_angles()):
        t1 = (_np.sin(i) * _np.sin(i) * _np.sin(2.0*W) * ( _np.arccos(-1.0/es) + _np.sqrt(es*es - 1.0) )).value
        t2 = ((1.0/3.0) * (1.0 + _np.cos(i)*_np.cos(i)) * _np.cos(2.0*w) * _np.sin(2.0*W)).value
        t3 = (2.0 * _np.cos(i) * _np.sin(2.0*w) * _np.cos(2.0*W) * ((es*es - 1.0)**1.5)/(es*es)).value

    if averaged:
        noncircular_result = (prefactor * ( ( _np.arccos(-1.0/es).value + _np.sqrt(es*es - 1.0) ) + (2.0/3.0) + (2.0 * ((es*es - 1.0)**1.5)/(es*es)) )).decompose(list(unit_set.values()))
    else: noncircular_result = (prefactor * (t1 + t2 + t3)).decompose(list(unit_set.values()))

    # Case: Circular Binary

    prefactor = (15.0/8.0) * (m3 * _np.abs(m1-m2) / (M12*M12)) * _np.sqrt(M12/M123) * ((a/q)**2.5) * 1.0 / (es*es*es * ((1.0 + es)**2.5))
    
    def f1(e):
        with _u.set_enabled_equivalencies(_u.dimensionless_angles()):
            return ((e**4.0) * _np.arccos(-1.0/e) + (-2.0 * 9.0*e*e + 8.0*e**4.0) * _np.sqrt(e*e - 1.0) / 15.0).to(_u.dimensionless_unscaled)
    with _u.set_enabled_equivalencies(_u.dimensionless_angles()):
        t1 = (_np.cos(i)**2.0 * _np.sin(w)**2.0) * (f1(es) * (1.0 - 3.75 * _np.sin(i)**2.0) + (2.0/15.0) * (es*es - 1.0)**2.5 * (1.0 - 5.0 * _np.sin(w)**2.0 * _np.sin(i)**2.0))**2.0
        t2 = (_np.cos(w)**2.0) * (f1(es) * (1.0 - 1.25 * _np.sin(i)**2.0) + (2.0/15.0) * (es*es - 1.0)**2.5 * (1.0 - 5.0 * _np.sin(w)**2.0 * _np.sin(i)**2.0))**2.0

# (187 (-1 + e^2)^5)/14400 + (199 f1[e] (4 (-1 + e^2)^(5/2) + 15 f1[e]))/7680
    if averaged: 
        with _u.set_enabled_equivalencies(_u.dimensionless_angles()):
            circular_result = prefactor * _np.sqrt(((187.0/14400.0)*(es*es - 1.0)**5.0) + ((199.0 * f1(es)) * (4.0 * (es*es - 1.0)**2.5 + 15.0 * f1(es)))/7680.0)
    else: circular_result = prefactor * _np.sqrt(t1 + t2)

    # Case: Exponential

    prefactor = (3.0 * _np.sqrt(2.0 * _np.pi)) * (m3 * (M12**0.25) / (M123**1.25)) * ((q/a)**0.75) * (((es + 1.0)**0.75)/(es*es))    
    with _u.set_enabled_equivalencies(_u.dimensionless_angles()):
        i2 = i/2.0
        exponential = -_np.sqrt(M12/M123) * ((q/a)**1.5) * ((_np.sqrt(es*es - 1.0) - _np.arccos(1.0/es)) / ((es - 1.0)**1.5))
        angles = (_np.cos(i2)**2.0) * _np.sqrt((_np.cos(i2)**4.0) + ((4.0/9.0)*_np.sin(i2)**4.0) + ((4.0/3.0)*(_np.cos(i2)**2.0)*(_np.sin(i2)**2.0)*_np.cos(4.0*w + 2.0*Wp)))

    if averaged: exponential_result = prefactor * _np.exp(exponential)
    else: exponential_result = prefactor * _np.exp(exponential) * angles

    # Add the ABS together, deliberately choose a norm.
    return (_np.nan_to_num(circular_result) + _np.nan_to_num(noncircular_result) + _np.nan_to_num(exponential_result)).decompose()

def parallel_eccentricity_change_adiabatic_estimate(sims, stars, averaged=False, particle_index=1, rmax=1e5*_u.au):
    """
    An analytical estimate for the changes in eccentricity of binary systems due to a flyby stars.

    Args:
        sims (list of Simulations): the simulation with two bodies, a central star and a planet.
        stars (Stars): the star or stars used to estimate the change in energy of a binary due to the flyby star
        averaged (bool, optional): boolean indicator for averaging over the incident angles of the flyby star. Default is False.
        particle_index (int, optional): the index of the particle in the simulation to calculate the energy for. Default is 1.
        rmax (float or Quantity, optional): needed to determine the time when the star will be at perihelion. Default is $10^5$ AU.

    Returns:
        results (list of Quantities): The changes in eccentricity of the binary systems.

    Example:
        ```python
        import rebound
        import airball
        def setup():
            sim = rebound.Simulation()
            sim.add(m=1.)
            sim.add(m=5e-5, a=30., e=0.1, f='uniform')
            return sim
        stars = airball.Star(m=1., b=500, v=5, size=10)
        sims = [setup() for _ in range(stars.N)]
        eccentricity_changes = airball.analytic.parallel_eccentricity_change_adiabatic_estimate(sims, stars)
        ```
    """
    return _joblib.Parallel(n_jobs=-1)(_joblib.delayed(eccentricity_change_adiabatic_estimate)(sim=sims[i], star=stars[i], averaged=averaged, particle_index=particle_index, rmax=rmax) for i in range(stars.N))

def eccentricity_change_impulsive_estimate(sim, star, particle_index=1):
    '''
    An analytical estimate for the change in eccentricity of an eccentric binary system due to a flyby star. From Equation (28) of [Heggie & Rasio (1996)](https://ui.adsabs.harvard.edu/abs/1996MNRAS.282.1064H/abstract). The orbital element angles of the flyby star are determined with respect to the plane defined by the binary orbit.
    
    The base assumptions are that the flyby is coplanar, $q_\star \gg a$, and $v_\star \gg v_\mathrm{planet}$. It may work outside of this regime, but it is not guaranteed.
    
    TODO:
        Make sure that the time of perihelion is calculated correctly and that the phase of the planet is correct.

    Args:
        sim (Simulation): The simulation with two bodies, a central star and a planet.
        star (Star): The star used to estimate the change in energy of a binary due to the flyby star
        particle_index (int, optional): The index of the particle in the simulation to calculate the energy for. Default is 1.

    Returns:
        result (Quantity): The change in eccentricity of the binary system.

    Example:
        ```python
        import rebound
        import airball
        sim = rebound.Simulation()
        sim.add(m=1.)
        sim.add(m=5e-5, a=30., e=0.1, inc=0.1)
        star = airball.Star(m=1., b=500, v=5)
        eccentricity_change = airball.analytic.eccentricity_change_impulsive_estimate(sim, star)
        ```
    '''

    index = int(particle_index)
    units = _tools.rebound_units(sim)
    G = (sim.G * units.length**3 / units.mass / units.time**2)

    p = sim.particles
    m1, m2, m3 = p[0].m * units.mass, p[index].m * units.mass, star.mass # redefine the masses for convenience
    M12 = m1 + m2 # total mass of the binary system
    
    a = p[index].a * units.length # redefine the orbital elements of the planet for convenience
    V = star.v
    q = star.q(sim) # compute the periapsis of the flyby star
    c = _tools.cartesian_elements(sim, star, rmax=0) # Get the heliocentric coordinates of the flyby star at periapsis
    dat = _np.array([c['x'].value, c['y'].value, c['z'].value]).T << units.length
    theta = _tools.angle_between(p[index].xyz, dat)

    # print(f"{G=}, {M12=}, {m3=}, {V=}, {a=}, {q=}, {theta=}")
    return ((2.0 * _np.sqrt(G/M12) * (m3/V) * _np.sqrt(a*a*a) / (q*q)) * _np.abs(_np.cos(theta)) * _np.sqrt((_np.cos(theta)**2.0) + (4.0 * _np.sin(theta)**2.0))).decompose()

def parallel_eccentricity_change_impulsive_estimate(sims, stars, particle_index=1):
    """
    An analytical estimate for the changes in eccentricity of binary systems due to a flyby stars.

    Args:
        sims (list of Simulations): the simulation with two bodies, a central star and a planet.
        stars (Stars): the star or stars used to estimate the change in energy of a binary due to the flyby star
        particle_index (int, optional): the index of the particle in the simulation to calculate the energy for. Default is 1.

    Returns:
        results (list of Quantities): The changes in eccentricity of the binary systems.

    Example:
        ```python
        import rebound
        import airball
        def setup():
            sim = rebound.Simulation()
            sim.add(m=1.)
            sim.add(m=5e-5, a=30., e=0.1, f='uniform')
            return sim
        stars = airball.Star(m=1., b=500, v=5, size=10)
        sims = [setup() for _ in range(stars.N)]
        eccentricity_changes = airball.analytic.parallel_eccentricity_change_impulsive_estimate(sims, stars)
        ```
    """
    return _joblib.Parallel(n_jobs=-1)(_joblib.delayed(eccentricity_change_impulsive_estimate)(sim=sims[i], star=stars[i], particle_index=particle_index) for i in range(stars.N))

############################################################
################ Inclination Estimate ######################
############################################################

def inclination_change_adiabatic_estimate(sim, star, averaged=False, particle_index=1):
    '''
    An analytical estimate for the sine of the change in inclination of an eccentric binary system due to a flyby star, $\\sin(\\Delta i)$. This function is based on Equation (A5) of [Malmberg, Davies, & Heggie (2011)](https://ui.adsabs.harvard.edu/abs/2011MNRAS.411..859M/abstract), with the addition of a factor of $(1 + e_\\star)^{-1/2}$ to account for hyperbolic flybys. The orbital element angles of the flyby star are determined with respect to the plane defined by the binary orbit.

    Args:
        sim (Simulation): the simulation with two bodies, a central star and a planet.
        star (Star or Stars): The star or stars that are flying by the binary system.
        averaged (bool, optional): Whether to return the inclination change averaged over the incident angles. Default is False.
        particle_index (int, optional): the index of the particle in the simulation to calculate the inclination change for. Default is 1.

    Returns:
        result (Quantity): Sine of the estimated change in the inclination, $\\sin(\\Delta i)$.

    Example:
        ```python
        import rebound
        import airball
        sim = rebound.Simulation()
        sim.add(m=1.)
        sim.add(m=5e-5, a=30., e=0.1, inc=0.1)
        star = airball.Star(m=1., b=500, v=5)
        inclination_change = airball.analytic.inclination_change_adiabatic_estimate(sim, star)
        ```
    '''

    index = int(particle_index)

    unit_set = _tools.rebound_units(sim)
    t0 = sim.t * unit_set['time']
    G = (sim.G * unit_set['length']**3 / unit_set['mass'] / unit_set['time']**2)
    
    p = sim.particles
    m1, m2, m3 = p[0].m * unit_set['mass'], p[index].m * unit_set['mass'], star.mass # redefine the masses for convenience
    M12 = m1 + m2 # total mass of the binary system
    M123 = m1 + m2 + m3 # total mass of all the objects involved
    
    mu = G * (_tools.system_mass(sim)  * unit_set['mass'] + m3)
    es = _tools.calculate_eccentricity(sim, star)
    
    a, e = p[index].a * unit_set['length'], p[index].e # redefine the orbital elements of the planet for convenience
    
    w, W, i = star.omega, star.Omega, star.inc # redefine the orientation elements of the flyby star for convenience

    V = star.v
    q = (- mu + _np.sqrt( mu**2. + star.b**2. * V**4.))/V**2. # compute the periapsis of the flyby star

    # Case: Non-Circular Binary

    prefactor = (3.0*_np.pi * m3 /8.0) * _np.sqrt(2.0/(M123 * M12 * (1.0 - e*e) * (1.0 + es))) * (a/q)**1.5
    with _u.set_enabled_equivalencies(_u.dimensionless_angles()):
        angles = (_np.sin(i) * _np.cos(i) * _np.sqrt( (1.0 + 3.0*e*e)**2.0 * _np.sin(W)**2.0 + (1.0 + e*e)**2.0 * _np.cos(W)**2.0 )).value

    if averaged:
        noncircular_result = (prefactor).decompose(list(unit_set.values()))
    else: noncircular_result = (prefactor * angles).decompose(list(unit_set.values()))

    return noncircular_result


import numpy as _np
import rebound as _rebound
import pickle as _pickle
from scipy.stats import uniform as _uniform
from . import environments as _env
from . import tools as _tools
from . import units as _u

try: from collections.abc import MutableMapping # Required for Python>=3.9
except: from collections import MutableMapping

class Star:
  '''
    This is the AIRBALL Star class.
    It encapsulates the relevant parameters for a given star. Only the mass is a quantity intrinsic to the object. The impact parameter, velocity, inclination, argument of periastron, and longitude of the ascending node quantities are defined with respect to the host star and plane passing through the star.

    Args:
      m (float): The mass of the star. Default units are in solMass.
      b (float): The impact parameter of the star. Default units are in AU. 
      v (float): The velocity at infinity of the star. Default units are in km/s.
      inc (float, optional): The inclination of the star. Default units are in radians.
      omega (float, optional): The argument of the periastron of the star. Default units are in radians.
      Omega (float, optional): The longitude of the ascending node of the star. Default units are in radians.
      UNIT_SYSTEM (list, optional): The unit system to use for the parameters. Default is [u.solMass, u.AU, u.km/u.s, u.rad].

    Attributes:
      m (astropy.units.Quantity): The mass of the star. Default units are in solMass.
      b (astropy.units.Quantity): The impact parameter of the star. Default units are in AU. 
      v (astropy.units.Quantity): The velocity at infinity of the star. Default units are in km/s.
      inc (astropy.units.Quantity): The inclination of the star. Default units are in radians.
      omega (astropy.units.Quantity): The argument of the periastron of the star. Default units are in radians.
      Omega (astropy.units.Quantity): The longitude of the ascending node of the star. Default units are in radians.
      units (airball.tools.UnitSet): The unit system to use for the parameters. Default is [u.solMass, u.AU, u.km/u.s, u.rad].
      impulse_gradient (astropy.units.Quantity): The impulse gradient of the star. Default units are in km/s/AU.
      params (list): A list of the parameters of the star in order of: Mass, m; Impact Parameter, b; Velocity, v; Inclination, inc; Argument of the Periastron, omega; and Longitude of the Ascending Node, Omega.
      param_values (list): A list of the values of the parameters of the star in order of: Mass, m; Impact Parameter, b; Velocity, v; Inclination, inc; Argument of the Periastron, omega; and Longitude of the Ascending Node, Omega.

    Examples:
      ```python
      import airball
      star = airball.Star(m=1.0, b=1.0, v=1.0, inc=0.0, omega=0.0, Omega=0.0)
      ```

      ```python
      import airball
      star = airball.Star(m=1.0, b=1.0, v=1.0, inc='uniform', omega='uniform', Omega='uniform')
      ```

      ```python
      import airball
      star = airball.Star(m=1.0, b=1.0, v=1.0)
      ```
  '''
  def __init__(self, m, b, v, inc='uniform', omega='uniform', Omega='uniform', UNIT_SYSTEM=[], **kwargs) -> None:
    self.units = _u.UnitSet(UNIT_SYSTEM)

    if inc == 'uniform' or inc == None: inc = 2.0*_np.pi * _uniform.rvs() - _np.pi
    if omega == 'uniform' or omega == None: omega = 2.0*_np.pi * _uniform.rvs() - _np.pi
    if Omega == 'uniform' or Omega == None: Omega = 2.0*_np.pi * _uniform.rvs() - _np.pi

    self.mass = m
    self.impact_parameter = b
    self.velocity = v
    self.inc = inc
    self.argument_periastron = omega
    self.longitude_ascending_node = Omega

  @property
  def UNIT_SYSTEM(self):
    return self.units.UNIT_SYSTEM

  @UNIT_SYSTEM.setter
  def UNIT_SYSTEM(self, UNIT_SYSTEM):
    self.units.UNIT_SYSTEM = UNIT_SYSTEM

  @property
  def N(self): return 1

  @property
  def m(self):
    return self._mass.to(self.units['mass'])

  @m.setter
  def m(self, value):
    self._mass = _tools.verify_unit(value, self.units['mass'])

  @property
  def mass(self):
    return self.m

  @mass.setter
  def mass(self, value):
    self.m = value
  
  @property
  def b(self):
    return self._impact_parameter.to(self.units['length'])

  @b.setter
  def b(self, value):
    self._impact_parameter = _tools.verify_unit(value, self.units['length'])

  @property
  def impact_parameter(self):
    return self.b

  @impact_parameter.setter
  def impact_parameter(self, value):
    self.b = value

  @property
  def v(self):
    return self._velocity.to(self.units['velocity'])

  @v.setter
  def v(self, value):
    self._velocity = _tools.verify_unit(value, self.units['velocity'])

  @property
  def velocity(self):
    return self.v

  @velocity.setter
  def velocity(self, value):
    self.v = value

  @property
  def inc(self):
    return self._inclination.to(self.units['angle'])

  @inc.setter
  def inc(self, value):
    self._inclination = _tools.verify_unit(value, self.units['angle'])

  @property
  def inclination(self):
    return self.inc

  @inc.setter
  def inclination(self, value):
    self.inc = value

  @property
  def omega(self):
    return self._argument_periastron.to(self.units['angle'])

  @omega.setter
  def omega(self, value):
    self._argument_periastron = _tools.verify_unit(value, self.units['angle'])

  @property
  def argument_periastron(self):
    return self.omega

  @argument_periastron.setter
  def argument_periastron(self, value):
    self.omega = value

  @property
  def Omega(self):
    return self._longitude_ascending_node.to(self.units['angle'])

  @Omega.setter
  def Omega(self, value):
    self._longitude_ascending_node = _tools.verify_unit(value, self.units['angle'])

  @property
  def longitude_ascending_node(self):
    return self.Omega

  @longitude_ascending_node.setter
  def longitude_ascending_node(self, value):
    self.Omega = value

  @property
  def impulse_gradient(self):
    # Calculate the impulse gradient for a flyby star.
    G = (1 * _u.au**3 / _u.solMass / _u.yr2pi**2)
    return ((2.0 * G * self.m) / (self.v * self.b**2.0)).to(_u.km/_u.s/_u.au)

  @property
  def params(self):
    # Returns a list of the parameters of the Stars (with units) in order of:
    # Mass, m; Impact Parameter, b; Velocity, v; Inclination, inc; Argument of the Periastron, omega; and Longitude of the Ascending Node, Omega
    return [self.m, self.b, self.v, self.inc, self.omega, self.Omega]
  
  @property
  def param_values(self):
    # Returns a list of the parameters of the Stars in order of:
    # Mass, m; Impact Parameter, b; Velocity, v; Inclination, inc; Argument of the Periastron, omega; and Longitude of the Ascending Node, Omega
    return _np.array([self.m.value, self.b.value, self.v.value, self.inc.value, self.omega.value, self.Omega.value])
   
  def eccentricity(self, sim):
    return self.e(sim)

  def e(self, sim):
    sim_units = _tools.rebound_units(sim)
    G = (sim.G * sim_units['length']**3 / sim_units['mass'] / sim_units['time']**2)
    mu = G * (_tools.system_mass(sim) * sim_units['mass'] + self.m)

    numerator = self.b * self.v*self.v
    return _np.sqrt(1 + (numerator/mu)**2.)
  
  def periastron(self, sim):
    return self.q(sim)

  def q(self, sim):
    sim_units = _tools.rebound_units(sim)
    G = (sim.G * sim_units['length']**3 / sim_units['mass'] / sim_units['time']**2)
    mu = G * (_tools.system_mass(sim)  * sim_units['mass'] + self.m)

    numerator = self.b * self.v*self.v
    star_e = _np.sqrt(1 + (numerator/mu)**2.)
    return self.b * _np.sqrt((star_e - 1.0)/(star_e + 1.0))
 
  def stats(self, returned=False):
    # Prints a summary of the current stats of the Star.
    s = f"<{self.__module__}.{type(self).__name__} object at {hex(id(self))}, "
    s += f"m= {self.mass:1.4g}, "
    s += f"b= {self.impact_parameter:1.4g}, "
    s += f"v= {self.velocity:1.4g}, "
    s += f"inc= {self.inc:1.4g}, "
    s += f"omega= {self.omega:1.4g}, "
    s += f"Omega= {self.Omega:1.4g}>"
    if returned: return s
    else: print(s)
  
  def __str__(self):
    return self.stats(returned=True)
  
  def __repr__(self):
    return self.stats(returned=True)
  
  def __len__(self):
    return NotImplemented
  
  def __eq__(self, other):
    # Overrides the default implementation
    if isinstance(other, Star):
        data = ((self.m == other.m) and (self.b == other.b) and (self.v == other.v) and (self.inc == other.inc) and (self.omega == other.omega) and (self.Omega == other.Omega))
        properties = (self.units == other.units)
        return data and properties
    return NotImplemented

  def __hash__(self):
    # Overrides the default implementation.
    data = []
    for d in sorted(self.__dict__.items()):
        try: data.append((d[0], tuple(d[1])))
        except: data.append(d)
    data = tuple(data)
    return hash(data)


class Stars(MutableMapping):
  '''
    This class allows the user to access stars like an array using the star's index. 
    Allows for negative indices and slicing.
    The implementation uses astropy.Quantity and numpy ndarrays underneath and only generates a airball.Star object when a single Star is requested.

    Args:
      m (list, ndarray, or Quantity): The masses of the stars. Default units are in solMass.
      b (list, ndarray, or Quantity): The impact parameters of the stars. Default units are in AU.
      v (list, ndarray, or Quantity): The velocities at infinity of the stars. Default units are in km/s.
      inc (list, ndarray, or Quantity, optional): The inclinations of the stars. Default units are in radians.
      omega (list, ndarray, or Quantity, optional): The arguments of the periastron of the stars. Default units are in radians.
      Omega (list, ndarray, or Quantity, optional): The longitudes of the ascending node of the stars. Default units are in radians.
      UNIT_SYSTEM (list, optional): The unit system to use for the parameters. Default is [u.solMass, u.AU, u.km/u.s, u.rad].
      size (int, optional): The number of stars to generate. Default is None.
      environment (airball.Environment, optional): The environment to generate the stars in (if `size` > 1). `env` is an alias. Default is None.
      filename (str, optional): The name of the file to load the instance from. The file should be in binary format. Default is None.

    Attributes:
      m (astropy.units.Quantity): The masses of the stars. Default units are in solMass.
      b (astropy.units.Quantity): The impact parameters of the stars. Default units are in AU.
      v (astropy.units.Quantity): The velocities at infinity of the stars. Default units are in km/s.
      inc (astropy.units.Quantity): The inclinations of the stars. Default units are in radians.
      omega (astropy.units.Quantity): The arguments of the periastron of the stars. Default units are in radians.
      Omega (astropy.units.Quantity): The longitudes of the ascending node of the stars. Default units are in radians.
      units (airball.tools.UnitSet): The unit system to use for the parameters. Default is [u.solMass, u.AU, u.km/u.s, u.rad].
      median_mass (astropy.units.Quantity): The median mass of the stars. Default units are in solMass.
      mean_mass (astropy.units.Quantity): The mean mass of the stars. Default units are in solMass.
      N (int): The number of stars.
      shape (tuple): The shape of the stars array.
      params (list): A list of the parameters of the stars in order of: Mass, m; Impact Parameter, b; Velocity, v; Inclination, inc; Argument of the Periastron, omega; and Longitude of the Ascending Node, Omega.
      param_values (list): A list of the values of the parameters of the stars in order of: Mass, m; Impact Parameter, b; Velocity, v; Inclination, inc; Argument of the Periastron, omega; and Longitude of the Ascending Node, Omega.

    Examples:
      ```python
      # Explicitly specify the stellar parameters.
      import airball
      stars_from_params1 = airball.Stars(m=[1.0, 1.0, 1.0], b=[1.0, 1.0, 1.0], v=[1.0, 1.0, 1.0], inc=[0.0, 0.0, 0.0], omega=[0.0, 0.0, 0.0], Omega=[0.0, 0.0, 0.0])
      stars_from_params2 = airball.Stars(m=[1.0, 1.0, 1.0], b=[1.0, 1.0, 1.0], v=[1.0, 1.0, 1.0]) # random orientation angles
      stars_from_params3 = airball.Stars(m=[1.0, 1.0, 1.0], b=[1.0, 1.0, 1.0], v=[1.0, 1.0, 1.0], inc='uniform', omega='uniform', Omega='uniform') # random orientation angles
      stars_from_params4 = airball.Stars(m=1, b=200, v=5, omega=0, Omega=0, size=100) # 100 identical stars with random inclinations
      ```

      ```python
      # Randomly generate the stellar parameters from a given environment.
      import airball
      stars_from_env1 = airball.Stars(environment=airball.OpenCluster(), size=3)
      stars_from_env2 = airball.Stars(env=airball.LocalNeighborhood(), size=5)
      stars_from_env3 = airball.Stars(airball.GlobularCluster(), size=9)
      ```

      ```python
      # Load the stellar parameters from a file.
      import airball
      stars_from_file1 = airball.Stars(filename='open_cluster.stars')
      stars_from_file2 = airball.Stars('open_cluster.stars')
      ```
  '''
  def __init__(self, filename=None, **kwargs) -> None:
    try: 
      self.units = _u.UnitSet(kwargs['UNIT_SYSTEM'])
      del kwargs['UNIT_SYSTEM']
    except KeyError:
      try: self.units = kwargs['units']
      except KeyError: 
        self.units = _u.UnitSet()

    # Initialize Stars from file.
    if filename is not None and isinstance(filename, str):
      try:
        loaded = Stars._load(filename)
        self.__dict__ = loaded.__dict__
      except: raise Exception('Invalid filename.')
      return
    
    # If nothing is specified, return an empty Stars object.
    if not kwargs: 
      self._m = _np.array([]) << self.units['mass']
      self._b = _np.array([]) << self.units['length']
      self._v = _np.array([]) << self.units['velocity']
      self._inc = _np.array([]) << self.units['angle']
      self._omega = _np.array([]) << self.units['angle']
      self._Omega = _np.array([]) << self.units['angle']
      self._shape = (0,)
      self.environment = None
      return

    # Determine the number of stars to generate.
    self._shape = (0,)
    for key in kwargs:
      try:
        if isinstance(kwargs[key], _u.Quantity):
            shape = kwargs[key].shape
            N = 0
            if len(shape) >= 1: N = _np.prod(shape) 
            else: shape = (0,)
            if N > self.N: self._shape = shape
        elif _tools.isList(kwargs[key]):
          if _np.prod(_np.shape(kwargs[key])) > self.N: self._shape = _np.shape(kwargs[key])
        else: pass
      except TypeError as err:
        print(err)
        self._shape = (len(kwargs[key]),)
      except Exception as err: raise err

    if 'size' in kwargs and self.N != 0: raise OverspecifiedParametersException('If lists are given then size cannot be specified.')
    elif 'size' in kwargs:
      if isinstance(kwargs['size'], tuple): 
        self._shape = tuple([int(i) for i in kwargs['size']])
      else: self._shape = (int(kwargs['size']),)
    elif self.N == 0: raise UnspecifiedParameterException('If no lists of parameters are given then size must be specified.')
    else: pass # No errors or issues, continue.


    # Initialize Stars from environment.
    self.environment = kwargs.get('environment', None)
    if (('environment' in kwargs or 'env' in kwargs) or  isinstance(filename, _env.StellarEnvironment)) and 'size' in kwargs:
      if self.N <= 1: raise InvalidValueForKeyException("If a generating environment is given then size must be greater than 1.")
      if 'environment' in kwargs:
        se = kwargs['environment'] 
        del kwargs['environment']
      elif 'env' in kwargs:
        se = kwargs['env']
        del kwargs['env']
      else:
        se = filename
        del filename
      stars = se.random_stars(**kwargs)
      self.__dict__ = stars.__dict__
      return

    # Initialize Stars from kwargs.
    keys = ['m', 'b', 'v']
    units = ['mass', 'length', 'velocity']
    unspecifiedParameterExceptions = ['Mass, m, must be specified.', 'Impact Parameter, b, must be specified.', 'Velocity, v, must be specified.']

    for k,u,upe in zip(keys, units, unspecifiedParameterExceptions):
      try:
        # Check to see if was key is given.
        value = kwargs[k]
        # Check if shape matches other key values.
        if len(value) != len(self): raise ListLengthException(f'Difference of {len(value)} and {len(self)} for {k}.')
        # Length of value matches other key values, check if value is a list.
        elif isinstance(value, list):
          # Value is a list, try to turn list of Quantities into a ndarray Quantity.
          try: quantityValue = _np.array([v.to(self.units[u]).value for v in value]) << self.units[u]
          # Value was not a list of Quantities, turn list into ndarray and make a Quantity.
          except: quantityValue = _np.array(value) << self.units[u]
        # Value was not a list, check to see if value is an ndarray.
        elif isinstance(value, _np.ndarray):
          # Assume ndarray is a Quantity and try to convert ndarray into given units.
          try: quantityValue = value.to(self.units[u])
          # ndarray is not a Quantity so turn ndarray into a Quantity.
          except: quantityValue = value << self.units[u]
        # Value implements __len__, but is not a list or ndarray.
        else: raise IncompatibleListException()
      # This key is necessary and must be specified, raise and Exception.
      except KeyError: raise UnspecifiedParameterException(upe)
      # Value is not a list, so assume it is an int or float and generate an ndarray of the given value.
      except TypeError: 
        value = value.to(self.units[u]) if _tools.isQuantity(value) else value * self.units[u]
        quantityValue = _np.ones(self.shape) * value
      # Catch any additional Exceptions.
      except Exception as err: raise err
      # Store Quantity Value as class property.
      if k == 'm': self._m = quantityValue
      elif k == 'b': self._b = quantityValue
      elif k == 'v': self._v = quantityValue
      else: raise _tools.InvalidKeyException()
      # Double check for consistent shapes.
      if self._shape is not None:
        if quantityValue.shape != self._shape: raise ListLengthException(f'Difference of {quantityValue.shape} and {self._shape} for {k}.')
      else: self._shape = quantityValue.shape
        
    for k in ['inc', 'omega', 'Omega']:
      try:
        # Check to see if was key is given.
        value = kwargs[k]
          # Check to see if value for key is string.
        if isinstance(value, str):
          # Value is a string, check to see if value for key is valid.
          if value != 'uniform': raise InvalidValueForKeyException()
          # Value 'uniform' for key is valid, now generate an array of values for key.
          quantityValue = (2.0*_np.pi * _uniform.rvs(size=self.shape) - _np.pi) * self.units['angle']
        # Value is not a string, check if length matches other key values.
        elif _np.shape(value) != self.shape: raise ListLengthException(f'Difference of {_np.shape(value)} and {self.shape} for {k}.')
        # Length of value matches other key values, check if value is a list.
        elif isinstance(value, list):
          # Value is a list, try to turn list of Quantities into a ndarray Quantity.
          try: quantityValue = _np.array([v.to(self.units['angle']).value for v in value]) * self.units['angle']
          # Value was not a list of Quantities, turn list into ndarray and make a Quantity.
          except: quantityValue = _np.array(value) * self.units['angle']
        # Value was not a list, check to see if value is an ndarray.
        elif isinstance(value, _np.ndarray):
          # Assume ndarray is a Quantity and try to convert ndarray into given units.
          try: quantityValue = value.to(self.units['angle'])
          # ndarray is not a Quantity so turn ndarray into a Quantity.
          except: quantityValue = value * self.units['angle']
        # Value implements __len__, but is not a list or ndarray.
        else: raise IncompatibleListException()
      # Key does not exist, assume the user wants an array of values to automatically be generated.
      except KeyError: 
        quantityValue = (2.0*_np.pi * _uniform.rvs(size=self.shape) - _np.pi) * self.units['angle']
      # Value is not a list, so assume it is an int or float and generate an ndarray of the given value.
      except TypeError: 
        value = value.to(self.units['angle']) if _tools.isQuantity(value) else value * self.units['angle']
        quantityValue = _np.ones(self.shape) * value
      # Catch any additional Exceptions.
      except Exception as err: raise err
      # Store Quantity Value as class property.
      if k == 'inc': self._inc = quantityValue
      elif k == 'omega': self._omega = quantityValue
      elif k == 'Omega': self._Omega = quantityValue
      else: raise _tools.InvalidKeyException()
      # Double check for consistent shapes.
      if self.shape is not None:
        if quantityValue.shape != self.shape: raise ListLengthException(f'Difference of {quantityValue.shape} and {self.shape} for {k}.')
      else: self._shape = quantityValue.shape

  @property
  def N(self):
    '''The total number of Stars.'''
    return _np.prod(self.shape)
    
  @property
  def shape(self):
    '''The shape of the Stars arrays.'''
    return self._shape

  @property
  def median_mass(self):
    '''The median mass of the Stars.'''
    return _np.median([mass.value for mass in self.m]) << self.units['mass']
  
  @property
  def mean_mass(self):
    '''The mean mass of the Stars.'''
    return _np.mean([mass.value for mass in self.m]) << self.units['mass']
  
  @property
  def m(self):
    return self._m << self.units['mass']
  
  @m.setter
  def m(self, value):
    try: 
      if _np.shape(value) != self.shape:  raise ListLengthException(f'Difference of {_np.shape(value)} and {self.shape}.')
    except (TypeError, AttributeError): raise IncompatibleListException()
    self._m = _tools.verify_unit(value, self.units['mass'])

  @property
  def mass(self):
    """
    Set the masses of the Stars, alias of `m` (default units: Msun).

    Args:
      value (list, ndarray, or Quantity): The new mass values. Can be set with a list such that len(value) == len(Stars).

    Raises:
      ListLengthException: If the length of the provided list does not match the number of Stars.
      IncompatibleListException: If the provided list is not compatible.
    """
    return self.m
  
  @mass.setter
  def mass(self, value):
    try: self.m = value
    except Exception as err: raise err
  
  @property
  def b(self):
    return self._b << self.units['length']
  
  @b.setter
  def b(self, value):
    try: 
      if _np.shape(value) != self.shape:  raise ListLengthException(f'Difference of {_np.shape(value)} and {self.shape}.')
    except (TypeError, AttributeError): raise IncompatibleListException()
    self._b = _tools.verify_unit(value, self.units['length'])

  @property
  def impact_parameter(self):
    """
    Set the impact parameters of the Stars, alias of `b` (default units: AU).

    Args:
      value (list, ndarray, or Quantity): The new impact parameter values. Can be set with a list such that len(value) == len(Stars).

    Raises:
      ListLengthException: If the length of the provided list does not match the number of Stars.
      IncompatibleListException: If the provided list is not compatible.
    """
    return self.b
  
  @impact_parameter.setter
  def impact_parameter(self, value):
    try: self.b = value
    except Exception as err: raise err

  @property
  def v(self):
    return self._v << self.units['velocity']

  @v.setter
  def v(self, value):
    try: 
      if _np.shape(value) != self.shape:  raise ListLengthException(f'Difference of {_np.shape(value)} and {self.shape}.')
    except (TypeError, AttributeError): raise IncompatibleListException()
    self._v = _tools.verify_unit(value, self.units['velocity'])

  @property
  def velocity(self):
    '''
    Set the velocities at infinity of the Stars, alias of `v` (default units: km/s).

    Args:
      value (list, ndarray, or Quantity): The new velocity values. Can be set with a list such that len(value) == len(Stars).

    Raises:
      ListLengthException: If the length of the provided list does not match the number of Stars.
      IncompatibleListException: If the provided list is not compatible.
    '''
    return self.v
  
  @velocity.setter
  def velocity(self, value):
    try: self.v = value
    except Exception as err: raise err

  @property
  def inc(self):
    return self._inc << self.units['angle']

  @inc.setter
  def inc(self, value):
    try: 
      if _np.shape(value) != self.shape:  raise ListLengthException(f'Difference of {_np.shape(value)} and {self.shape}.')
    except (TypeError, AttributeError): raise IncompatibleListException()
    self._inc = _tools.verify_unit(value, self.units['angle'])

  @property
  def inclination(self):
    '''
    The inclinations of the Stars, alias of `inc` (default units: radians).

    Args:
      value (list, ndarray, or Quantity): The new inclination values. Can be set with a list such that len(value) == len(Stars).

    Raises:
      ListLengthException: If the length of the provided list does not match the number of Stars.
      IncompatibleListException: If the provided list is not compatible.
    '''
    return self.inc
  
  @inclination.setter
  def inclination(self, value):
    try: self.inc = value
    except Exception as err: raise err

  @property
  def omega(self):
    return self._omega << self.units['angle']

  @omega.setter
  def omega(self, value):
    try: 
      if _np.shape(value) != self.shape:  raise ListLengthException(f'Difference of {_np.shape(value)} and {self.shape}.')
    except (TypeError, AttributeError): raise IncompatibleListException()
    self._omega = _tools.verify_unit(value, self.units['angle'])

  @property
  def argument_periastron(self):
    '''
    The argument of periastron of the Stars, alias of `omega` (default units: radians).

    Args:
      value (list, ndarray, or Quantity): The new values for the argument of periastron. Can be set with a list such that len(value) == len(Stars).

    Raises:
      ListLengthException: If the length of the provided list does not match the number of Stars.
      IncompatibleListException: If the provided list is not compatible.
    '''
    return self.omega

  @argument_periastron.setter
  def argument_periastron(self, value):
    try: self.omega = value
    except Exception as err: raise err

  @property
  def Omega(self):
    return self._Omega << self.units['angle']

  @Omega.setter
  def Omega(self, value):
    try: 
      if _np.shape(value) != self.shape:  raise ListLengthException(f'Difference of {_np.shape(value)} and {self.shape}.')
    except (TypeError, AttributeError): raise IncompatibleListException()
    self._Omega = _tools.verify_unit(value, self.units['angle'])

  @property
  def longitude_ascending_node(self):
    '''
    The longitude of the ascending node of the Stars, alias of `Omega` (default units: radians).

    Args:
      value (list, ndarray, or Quantity): The new longitude of the ascending node values. Can be set with a list such that len(value) == len(Stars).

    Raises:
      ListLengthException: If the length of the provided list does not match the number of Stars.
      IncompatibleListException: If the provided list is not compatible.
    '''
    return self.Omega
  
  @longitude_ascending_node.setter
  def longitude_ascending_node(self, value):
    try: self.Omega = value
    except Exception as err: raise err

  @property
  def impulse_gradient(self):
    '''
      Calculate the impulse gradient for a flyby star.
      
      $$\\frac{2GM_\\star}{V_\\star b_\\star^2}$$  
    '''
    G = (1 * _u.au**3 / _u.solMass / _u.yr2pi**2)
    return ((2.0 * G * self.m) / (self.v * self.b**2.0)).to(_u.km/_u.s/_u.au)

  @property
  def params(self):
    '''
      A list of the parameters of the Stars (with units) in order of: Mass, m; Impact Parameter, b; Velocity, v; Inclination, inc; Argument of the Periastron, omega; and Longitude of the Ascending Node, Omega

      `[m, b, v, inc, omega, Omega]`
    '''
    return [self.m, self.b, self.v, self.inc, self.omega, self.Omega]
  
  @property
  def param_values(self):
    '''
      A list of the values of the parameters of the Stars in order of: Mass, m; Impact Parameter, b; Velocity, v; Inclination, inc; Argument of the Periastron, omega; and Longitude of the Ascending Node, Omega

      `[m.value, b.value, v.value, inc.value, omega.value, Omega.value]`
    '''
    return _np.array([self.m.value, self.b.value, self.v.value, self.inc.value, self.omega.value, self.Omega.value])
 
  def eccentricity(self, sim):
    '''
    The eccentricity of the Stars, alias for `e(sim)`.

    Args:
      sim (Simulation): The REBOUND Simulation to use for calculating the eccentricity.

    Returns:
      eccentricity (ndarray): The eccentricity of the Stars.
    '''
    return self.e(sim)

  def e(self, sim):
    units = _tools.rebound_units(sim)
    G = (sim.G * units.length**3 / units.mass / units.time**2)
    mu = G * (_tools.system_mass(sim) * units.mass + self.m)

    numerator = self.b * self.v*self.v
    return _np.sqrt(1 + (numerator/mu)**2.)
  
  def periastron(self, sim):
    '''
    The periastron of the Stars, alias for `q(sim)`.

    Args:
      sim (Simulation): The REBOUND Simulation to use for calculating the periastron.

    Returns:
      periastron (ndarray): The periastron of the Stars.
    '''
    return self.q(sim)

  def q(self, sim):
    sim_units = _tools.rebound_units(sim)
    G = (sim.G * sim_units['length']**3 / sim_units['mass'] / sim_units['time']**2)
    mu = G * (_tools.system_mass(sim)  * sim_units['mass'] + self.m)

    numerator = self.b * self.v*self.v
    star_e = _np.sqrt(1 + (numerator/mu)**2.)
    return self.b * _np.sqrt((star_e - 1.0)/(star_e + 1.0))
  
  def copy(self):
    '''Returns a deep copy of the Stars.'''
    newstars = self[:]
    newstars.environment = self.environment.copy()
    return newstars

  def sort(self, key, sim=None, argsort=False):
    # Alias for `sortby`.
    return self.sortby(key, sim, argsort)
  
  def argsort(self, key, sim=None):
    # Alias for `sortby(argsort=True)`.
    return self.sortby(key, sim, argsort=True)
  
  def argsortby(self, key, sim=None):
    # Alias for `sortby(argsort=True)`.
    return self.sortby(key, sim, argsort=True)

  def sortby(self, key, sim=None, argsort=False):
    """
    Sort the Stars in ascending order by a defining parameter.

    Args:
      key (str): The parameter to sort by. Options include: 'm' (mass), 'b' (impact parameter), 'v' (relative velocity at infinity), 'inc' (inclination), 'omega' (argument of the periastron), 'Omega' (longitude of the ascending node), 'q' (periapsis), 'e' (eccentricity). For 'q' and 'e', a REBOUND Simulation is required.
      sim (Simulation, optional): The REBOUND Simulation to use for sorting by 'q' and 'e'. Default is None.
      argsort (bool, optional): If True, return the indices used to sort the Stars. Default is False.

    Returns:
      indices (list or None): The indices used to sort the Stars, if argsort is True. Otherwise, None and sorting is done in place.

    !!! Info
        The Stars can also be sorted arbitrarily by providing a list of indices of length len(stars) as the key.

    Examples:
      ```python
      import airball
      stars = airball.Stars(m=[1, 2, 3], b=[3,2,1], v=[6,5,7])
      stars.sortby('b')
      inds  = stars.sortby('v', argsort=True)
      ```

      ```python
      import airball
      import rebound
      sim = rebound.Simulation()
      sim.add(m=1)
      sim.add(m=5e-5, a=30, e=0.01)
      stars = airball.Stars(airball.OpenCluster(), size=100)
      stars.sortby('q', sim=sim)
      ```   

      ```python
      import airball
      stars = airball.Stars(airball.LocalNeighborhood(), size=5)
      inds = [0,4,1,3,2]
      stars.sortby(inds)
      ```      
    """

    inds = _np.arange(len(self))
    if key == 'm' or key == 'mass': inds = _np.argsort(self.m)
    elif key == 'b' or key == 'impact' or key == 'impact param' or key == 'impact parameter': inds = _np.argsort(self.b)
    elif key == 'v' or key == 'vinf' or key == 'v_inf' or key == 'velocity': inds = _np.argsort(self.v)
    elif key == 'inc' or key == 'inclination' or key == 'i' or key == 'I': inds = _np.argsort(self.inc)
    elif key == 'omega' or key == 'ω' or key == 'argument_periastron': inds = _np.argsort(self.omega)
    elif key == 'Omega' or key == 'Ω' or key == 'longitude_ascending_node': inds = _np.argsort(self.Omega)
    elif key == 'q' or key == 'peri' or key == 'perihelion' or key == 'periastron' or key == 'periapsis': 
      if isinstance(sim, _rebound.Simulation):
        inds = _np.argsort(self.q(sim))
      else: raise InvalidParameterTypeException()
    elif key == 'e' or key == 'eccentricity': 
      if isinstance(sim, _rebound.Simulation):
        inds = _np.argsort(self.e(sim))
      else: raise InvalidParameterTypeException()
    elif _tools.isList(key): 
      print(key)
      if len(key) != len(self): raise ListLengthException(f'Difference of key: {len(key)} and stars: {len(self)}.')
      inds = _np.array(key)
    else: raise InvalidValueForKeyException()

    if argsort: return inds
    else:
      self.m[:] = self.m[inds]
      self.b[:] = self.b[inds]
      self.v[:] = self.v[inds]
      self.inc[:] = self.inc[inds]
      self.omega[:] = self.omega[inds]
      self.Omega[:] = self.Omega[inds]

  def save(self, filename):
    """
    Save the current instance of the Stars class to a file using pickle.

    Args:
      filename (str): The name of the file to save the instance to. The file will be saved in binary format.

    Example:
      ```python
      import airball
      se = airball.OpenCluster()
      stars = se.random_stars(100)
      stars.save('open_cluster.stars')
      ```
    """
    if not isinstance(filename, str): raise ValueError('Filename must be a string.')
    with open(filename, 'wb') as pfile:
      _pickle.dump(self, pfile, protocol=_pickle.HIGHEST_PROTOCOL)

  @classmethod
  def _load(cls, filename, init=False):
    """
    Load an instance of the Stars class from a file using pickle.

    Args:
      filename (str): The name of the file to load the instance from. The file should be in binary format, pickled.

    Returns:
      loaded_stars (Stars): The loaded instance of the Stars class.

    Example:
      ```python
      import airball
      stars = airball.Stars('open_cluster.stars')
      ```
    """
    if not isinstance(filename, str): raise ValueError('Filename must be a string.')
    return _pickle.load(open(filename, 'rb'))

  def stats(self, returned=False):
    '''
    Prints a summary of the current stats of the Stars object.
    The stats are returned as a string if `returned=True`.
    '''
    s = f"<{self.__module__}.{type(self).__name__} object at {hex(id(self))}, "
    s += f"N={f'{self.N:,.0f}' if len(self.shape) == 1 else self.shape}"
    if self.N > 0: s += f", m= {_np.min(self.m.value):,.2f}-{_np.max(self.m.value):,.2f} {self.units['mass']}"
    if self.N > 0: s += f", b= {_np.min(self.b.value):,.0f}-{_np.max(self.b.value):,.0f} {self.units['length']}"
    if self.N > 0: s += f", v= {_np.min(self.v.value):,.0f}-{_np.max(self.v.value):,.0f} {self.units['velocity']}"
    s += f"{f', Environment={self.environment.name}' if self.environment is not None else ''}"
    s += ">"
    if returned: return s
    else: print(s)
  
  def __str__(self):
    return self.stats(returned=True)
  
  def __repr__(self):
    return self.stats(returned=True)

  def __getitem__(self, key):
    int_types = int, _np.integer
  
    # Basic indexing.
    if isinstance(key, int_types):
      # If the set of Stars is multi-dimensional, return the requested subset of stars as a set of Stars.
      if len(self.m.shape) > 1: return Stars(m=self.m[key], b=self.b[key], v=self.v[key], inc=self.inc[key], omega=self.omega[key], Omega=self.Omega[key], UNIT_SYSTEM=self.units.UNIT_SYSTEM)
      # Otherwise return the requested Star.
      else: return Star(m=self.m[key], b=self.b[key], v=self.v[key], inc=self.inc[key], omega=self.omega[key], Omega=self.Omega[key], UNIT_SYSTEM=self.units.UNIT_SYSTEM)

    # Allows for boolean array masking and indexing using a subset of indices.
    if isinstance(key, _np.ndarray):
      return Stars(m=self.m[key], b=self.b[key], v=self.v[key], inc=self.inc[key], omega=self.omega[key], Omega=self.Omega[key], UNIT_SYSTEM=self.units.UNIT_SYSTEM)
    
    # Allow for speed efficient slicing by returning a new set of Stars which are a subset of the original object.
    if isinstance(key, slice):
      # Check for number of elements returned by the slice.
      numEl = _tools.numberOfElementsReturnedBySlice(*key.indices(self.N))
            # If the slice requests the entire set, then simply return the set.
            # if key == slice(None, None, None): return self #  !!! **Note: this is a reference to the same object.** !!!
      # If there are no elements requested, return the empty set.
      if numEl == 0: return Stars(m=[], b=[], v=[], size=0)
      # If only one element is requested, return a set of Stars with only one Star.
      elif numEl == 1: return Stars(m=self.m[key], b=self.b[key], v=self.v[key], inc=self.inc[key], omega=self.omega[key], Omega=self.Omega[key], UNIT_SYSTEM=self.units.UNIT_SYSTEM, size=1)
      # Otherwise return a subset of the Stars defined by the slice.
      else: return Stars(m=self.m[key], b=self.b[key], v=self.v[key], inc=self.inc[key], omega=self.omega[key], Omega=self.Omega[key], UNIT_SYSTEM=self.units.UNIT_SYSTEM)

    # Allow for Numpy style array indexing.
    if isinstance(key, tuple):
      # Check if Stars data is multi-dimensional.
      if len(self.m.shape) == 1: raise IndexError(f'Too many indices: Stars are 1-dimensional, but {len(key)} were indexed.')
      # Check to see if the tuple has a slice.
      hasSlice = _tools.hasTrue([isinstance(k, slice) for k in key])
      if hasSlice:
        # Check the number of elements requested by the slice.
        numEl = [_tools.numberOfElementsReturnedBySlice(*k.indices(self.m.shape[i])) if isinstance(k, slice) else 1 for i,k in enumerate(key)]
        # If there are no elements requested, return the empty set.
        if numEl.count(0) > 0: return Stars(m=[], b=[], v=[], size=0)
        # If multiple elements are requested, return a set of Stars.
        elif _np.any(_np.array(numEl) > 1): return Stars(m=self.m[key], b=self.b[key], v=self.v[key], inc=self.inc[key], omega=self.omega[key], Omega=self.Omega[key], UNIT_SYSTEM=self.units.UNIT_SYSTEM)
        # If only one element is requested, return a set of Stars with only one Star.
        else:
          # Check to see if the single element is an scalar or an array with only one element.
          if self.m[key].isscalar: return Stars(m=self.m[key], b=self.b[key], v=self.v[key], inc=self.inc[key], omega=self.omega[key], Omega=self.Omega[key], size=1, UNIT_SYSTEM=self.units.UNIT_SYSTEM)
          else: return Stars(m=self.m[key], b=self.b[key], v=self.v[key], inc=self.inc[key], omega=self.omega[key], Omega=self.Omega[key], UNIT_SYSTEM=self.units.UNIT_SYSTEM)
      # If there is no slice, the return the requested Star.
      else: return Star(m=self.m[key], b=self.b[key], v=self.v[key], inc=self.inc[key], omega=self.omega[key], Omega=self.Omega[key], UNIT_SYSTEM=self.units.UNIT_SYSTEM)

    raise _tools.InvalidKeyException() 

  def __setitem__(self, key, value):
    star_type = Star, Stars
    if isinstance(value, star_type):
      self.m[key], self.b[key], self.v[key], self.inc[key], self.omega[key], self.Omega[key] = value.params
    else: raise InvalidStarException()
  
  def __delitem__(self, key):
    raise ValueError('Cannot delete Star elements from Stars array.')

  def __iter__(self):
    for i in list(_np.ndindex(self.m.shape)):
      yield Star(m=self.m[i], b=self.b[i], v=self.v[i], inc=self.inc[i], omega=self.omega[i], Omega=self.Omega[i], UNIT_SYSTEM=self.units.UNIT_SYSTEM)

  def __len__(self):
    return self.shape[0]
  
  def __eq__(self, other):
    # Overrides the default implementation
    if isinstance(other, Stars):
        data = (_np.all(self.m == other.m) and _np.all(self.b == other.b) and _np.all(self.v == other.v) and _np.all(self.inc == other.inc) and _np.all(self.omega == other.omega) and _np.all(self.Omega == other.Omega))
        properties = (self.N == other.N and self.shape == self.shape and self.units == other.units and self.environment == other.environment)
        return data and properties
    return NotImplemented

  def __hash__(self):
    # Overrides the default implementation
    data = []
    for d in sorted(self.__dict__.items()):
        try: data.append((d[0], tuple(d[1])))
        except: data.append(d)
    data = tuple(data)
    return hash(data)
  
  def __add__(self, other):
    # Overrides the default implementation
    if isinstance(other, Stars):
      if self.N == 0 and other.N == 0: return Stars()
      else: return Stars(m=_np.concatenate((self.m, other.m)), b=_np.concatenate((self.b, other.b)), v=_np.concatenate((self.v, other.v)), inc=_np.concatenate((self.inc, other.inc)), omega=_np.concatenate((self.omega, other.omega)), Omega=_np.concatenate((self.Omega, other.Omega)))
    return NotImplemented

################################
###### Custom Exceptions #######
################################

class InvalidStarException(Exception):
  def __init__(self): super().__init__('Object is not a valid airball.Star object.')

class UnspecifiedParameterException(Exception):
  def __init__(self, message): super().__init__(message)

class OverspecifiedParametersException(Exception):
  def __init__(self, message): super().__init__(message)

class ListLengthException(Exception):
  def __init__(self, message): super().__init__(f'List arguments must be same length or shape. {message}')

class IncompatibleListException(Exception):
  def __init__(self): super().__init__('The given list type is not compatible. Please use a Python list or an ndarray.')

class InvalidValueForKeyException(Exception):
  def __init__(self, message=''): super().__init__(f'Invalid value for key. {message}')

class InvalidParameterTypeException(Exception):
  def __init__(self): super().__init__('The given parameter value is not a valid type.')




import astropy.units as _u
from astropy.units import *
twopi = 6.28318530717958623199592693708837032318115234375
yrtwopi = _u.def_unit('yrtwopi', _u.yr/twopi, format={'latex': r'(yr/2\pi)'})
yr2pi = _u.def_unit('yr2pi', _u.yr/twopi, format={'latex': r'(yr/2\pi)'})
stars = _u.def_unit('stars')
_u.add_enabled_units([yr2pi, yrtwopi])
_u.add_enabled_aliases({'msun': _u.solMass})


def isUnit(var):
    '''Determines if an object is an Astropy Quantity. Used for Stellar Environment initializations.'''
    return isinstance(var, (_u.core.IrreducibleUnit, _u.core.CompositeUnit, _u.Unit))

class UnitSet():
  '''
    A class for managing the units of another class.
    The default units are `[u.au, u.yr2pi, u.solMass, u.rad, u.km/u.s, u.stars/u.pc**3, u.stars]`.
    If a list of units is provided, the UnitSet will attempt to determine the units of length, time, mass, angle, velocity, density, and object. Thus, if a unit of length and a unit of time are given, but no unit of velocity, then UnitSet will calculate the unit of velocity from the given units of length and time. If no units are provided, the default units will be used.
    Two UnitSets are considered equal if the string representations of the units in each UnitSets are identical.

    Args:
      UNIT_SYSTEM (list): A list of Astropy Units describing the units of the system.

    Attributes:
      UNIT_SYSTEM (list): A list of Astropy Units describing the units of the system.
      units (dict): A dictionary of Astropy Units describing the units of the system. Can also access the dictionary from the object itself.
      length (astropy.units.Unit): The unit of length.
      time (astropy.units.Unit): The unit of time.
      mass (astropy.units.Unit): The unit of mass.
      angle (astropy.units.Unit): The unit of angle.
      velocity (astropy.units.Unit): The unit of velocity in length/time.
      density (astropy.units.Unit): The unit of density in object/length**3.
      object (airball.units.Unit): The unit of an object (such as a star).

    Example:
      ```python
      import airball
      import airball.units as u
      us1 = airball.tools.UnitSet([u.pc, u.Myr])
      us2 = airball.tools.UnitSet()
      print(us1 == us2) # False
      print(us1.velocity) # pc/Myr
      print(us2['velocity']) # au/yr2pi
      ```
  '''

  def __init__(self, UNIT_SYSTEM=[]) -> None:
    self._units = {'length': _u.au, 'time': _u.Myr, 'mass': _u.solMass, 'angle': _u.rad, 'velocity': _u.km/_u.s, 'object': stars, 'density': stars/_u.pc**3}
    self.UNIT_SYSTEM = UNIT_SYSTEM
    pass

  @property
  def units(self):
    return self._units

  @property
  def UNIT_SYSTEM(self):
    return self._UNIT_SYSTEM

  def __getitem__(self, key):
    if isinstance(key, str): return self.units[key]
    else: raise InvalidKeyException()

  def __setitem__(self, key, value):
    if isinstance(key, str):
      if isUnit(value): self.units[key] = value
      else: raise InvalidUnitException()
    else: raise InvalidKeyException()

  def __str__(self):
    s = '{'
    for key in self.units:
       s += f'{key}: {self.units[key].to_string()}, '
    s = s[:-2] + '}'
    return s

  def __repr__(self):
    s = '{'
    for key in self.units:
       s += f'{key}: {self.units[key].to_string()},\n'
    s = s[:-2] + '}'
    return s

  def __iter__(self):
    for k in self.units:
      yield self.units[k]

  def __eq__(self, other):
    # Determines if the string representations of the units in each UnitSets are identical.
    if isinstance(other, UnitSet):
        result = True
        for u1,u2 in zip(self, other):
            result = result and (u1.to_string() == u2.to_string())
        return result
    return NotImplemented

  def __hash__(self):
    # Overrides the default implementation
    data = []
    for d in sorted(self.__dict__.items()):
        try: data.append((d[0], tuple(d[1])))
        except: data.append(d)
    data = tuple(data)
    return hash(data)

  def values(self):
    return self.units.values()

  @property
  def length(self):
    return self._units['length']

  @length.setter
  def length(self, value):
    self.UNIT_SYSTEM = [value]

  @property
  def time(self):
    return self._units['time']

  @time.setter
  def time(self, value):
    self.UNIT_SYSTEM = [value]

  @property
  def mass(self):
    return self._units['mass']

  @mass.setter
  def mass(self, value):
    self.UNIT_SYSTEM = [value]

  @property
  def angle(self):
    return self._units['angle']

  @angle.setter
  def angle(self, value):
    self.UNIT_SYSTEM = [value]

  @property
  def velocity(self):
    return self._units['velocity']

  @velocity.setter
  def velocity(self, value):
    self.UNIT_SYSTEM = [value]

  @property
  def density(self):
    return self._units['density']

  @density.setter
  def density(self, value):
    self.UNIT_SYSTEM = [value]

  @property
  def object(self):
    return self._units['object']

  @object.setter
  def object(self, value):
    self.UNIT_SYSTEM = [value]

  @UNIT_SYSTEM.setter
  def UNIT_SYSTEM(self, UNIT_SYSTEM):
    if UNIT_SYSTEM != []:
      lengthUnit = [this for this in UNIT_SYSTEM if this.is_equivalent(_u.m)]
      self._units['length'] = lengthUnit[0] if lengthUnit != [] else self._units['length']

      timeUnit = [this for this in UNIT_SYSTEM if this.is_equivalent(_u.s)]
      self._units['time'] = timeUnit[0] if timeUnit != [] else self._units['time']

      velocityUnit = [this for this in UNIT_SYSTEM if this.is_equivalent(_u.km/_u.s)]
      if velocityUnit == [] and timeUnit != [] and lengthUnit != []: velocityUnit = [lengthUnit[0]/timeUnit[0]]
      self._units['velocity'] = velocityUnit[0] if velocityUnit != [] else self._units['velocity']

      massUnit = [this for this in UNIT_SYSTEM if this.is_equivalent(_u.kg)]
      self._units['mass'] = massUnit[0] if massUnit != [] else self._units['mass']

      angleUnit = [this for this in UNIT_SYSTEM if this.is_equivalent(_u.rad)]
      self._units['angle'] = angleUnit[0] if angleUnit != [] else self._units['angle']

      objectUnit = [this for this in UNIT_SYSTEM if this.is_equivalent(stars)]
      self._units['object'] = objectUnit[0] if objectUnit != [] else stars

      densityUnit = [this for this in UNIT_SYSTEM if this.is_equivalent(stars/_u.m**3)]
      densityUnit2 = [this for this in UNIT_SYSTEM if this.is_equivalent(1/_u.m**3)]
      if densityUnit == [] and densityUnit2 != []:
        densityUnit = [self._units['object'] * densityUnit2[0]]
      elif densityUnit == [] and objectUnit != [] and lengthUnit != []:
        densityUnit = [self._units['object']/self._units['length']**3]
      elif densityUnit == [] and densityUnit2 == [] and objectUnit != []:
         densityLength = [this for this in self._units['density'].bases if this.is_equivalent(_u.m)][0]
         densityUnit = [self._units['object']/densityLength**3]
      self._units['density'] = densityUnit[0] if densityUnit != [] else self._units['density']

    self._UNIT_SYSTEM = list(self._units.values())

############################################################
###################### Exceptions ##########################
############################################################

class InvalidKeyException(Exception):
  def __init__(self): super().__init__('Invalid key type.')

class InvalidUnitException(Exception):
  def __init__(self): super().__init__('Value is not a valid unit type.')



import rebound as _rebound
import numpy as _np
from scipy.stats import uniform as _uniform
from scipy.stats import maxwell as _maxwell
from scipy.stats import expon as _exponential
from scipy.optimize import fminbound as _fminbound

from . import units as _u
from . import imf as _imf
from .imf import IMF as _IMF
from .stars import Star as _Star
from .stars import Stars as _Stars
from . import analytic as _analytic
from . import tools as _tools

class StellarEnvironment:
  '''
    This is the AIRBALL StellarEnvironment class.
    It encapsulates the relevant data for a static stellar environment.

    Initializing a StellarEnvironment instance.

    Args:
      stellar_density (float): The stellar density of the environment.
      velocity_dispersion (float): The velocity dispersion in the environment.
      lower_mass_limit (float): The lower mass limit for stars in the environment.
      upper_mass_limit (float): The upper mass limit for stars in the environment.
      mass_function (callable, optional): A function that defines the mass distribution of stars. Default is None.
      maximum_impact_parameter (float, optional): The maximum impact parameter defining the outer limit of the sphere of influence around a stellar system. If not provided, AIRBALL attempts to estimate a reasonable one. Default is None.
      name (str, optional): The name of the environment. Default is None.
      UNIT_SYSTEM (list, optional): The unit system used in the environment. Default is an empty list.
      units (airball.units.UnitSet, optional): The units used in the environment. Default is None.
      object_name (str, optional): The name of the object in the environment. Default is None.
      seed (int, optional): The seed fixing the random star generator. Default is None so it's always random.

    Example:
      ```python
      import airball
      my_env = airball.StellarEnvironment(stellar_density=10, velocity_dispersion=20, lower_mass_limit=0.08, upper_mass_limit=8, name='My Environment')
      my_star = my_env.random_star()
      ```

    If a `maximum_impact_parameter` is not given, AIRBALL attempts to estimate a reasonable one.
    The Maximum Impact Parameter is radius defining the outer limit of the sphere of influence around a stellar system.
    There are predefined subclasses for the LocalNeighborhood, a generic OpenCluster, a generic GlobularCluster, and the Milky Way center GalacticBulge and GalacticCore.
  '''
  def __init__(self, stellar_density, velocity_dispersion, lower_mass_limit, upper_mass_limit, mass_function=None, maximum_impact_parameter=None, name=None, UNIT_SYSTEM=[], units=None, object_name=None, seed=None, number_imf_samples=100):
    # Check to see if an stars object unit is defined in the given UNIT_SYSTEM and if the user defined a different name for the objects.
    self.units = units if isinstance(units, _u.UnitSet) else _u.UnitSet(UNIT_SYSTEM)
    objectUnit = [this for this in UNIT_SYSTEM if this.is_equivalent(_u.stars)]
    if objectUnit == [] and object_name is not None: self.units.object = _u.def_unit(object_name, _u.stars)
    elif objectUnit == [] and object_name is None: self.units.object = _u.stars
    else: self.units.object = objectUnit[0]

    self.density = stellar_density
    self.velocity_dispersion = velocity_dispersion

    self._upper_mass_limit = upper_mass_limit.to(self.units['mass']) if _tools.isQuantity(upper_mass_limit) else upper_mass_limit * self.units['mass']
    self._lower_mass_limit = lower_mass_limit.to(self.units['mass']) if _tools.isQuantity(lower_mass_limit) else lower_mass_limit * self.units['mass']
    self._IMF = _IMF(min_mass=self._lower_mass_limit, max_mass=self._upper_mass_limit, mass_function=mass_function, unit=self.units['mass'], number_samples=number_imf_samples, seed=seed)
    self._median_mass = self.IMF.median_mass
    self.maximum_impact_parameter = maximum_impact_parameter

    self.name = name if name is not None else 'Stellar Environment'
    self.short_name = self.name.replace(' ', '')
    self.seed = seed if seed is not None else None #_np.random.randint(0, int(2**32 - 1))

  def random_stars(self, size=1, **kwargs):
    '''
      Computes a random star from a stellar environment.

      Args:
        size (int or tuple): The number of stars to generate. If size is a tuple, it is interpreted as array dimensions. Default: 1.

      Keyword Args:
        include_orientation (bool, optional): If True, the orientation of the star is randomly generated. Otherwise, the orientation of the stars are zero. Default: True.
        maximum_impact_parameter (float, optional): The maximum impact parameter of the star. If None, the maximum impact parameter is estimated. Default: None.
        seed (int, optional): The random seed to use. If None is given then it is random every time. Default: None.

      Returns:
        stars (Star or Stars): A Star object or Stars object (if size > 1) with the randomly generated masses, impact parameters, velocities, and orientations in a heliocentric model.

      Example:
        ```python
        import airball
        my_env = airball.StellarEnvironment(stellar_density=10, velocity_dispersion=20, lower_mass_limit=0.08, upper_mass_limit=8, name='My Environment')
        my_stars = my_env.random_stars(10)
        ```
    '''
    if isinstance(size, tuple): size = tuple([int(i) for i in size])
    else: size = int(size)

    include_orientation = kwargs.get('include_orientation', True)
    maximum_impact_parameter = kwargs.get('maximum_impact_parameter')
    seed = kwargs.get('seed')

    self.seed = seed
    if self.seed != None: _np.random.seed(self.seed)

    v = _maxwell.rvs(scale=_tools.maxwell_boltzmann_scale_from_dispersion(self.velocity_dispersion), size=size) << self.units['velocity'] # Relative velocity of the star at infinity.

    max_impact = maximum_impact_parameter if maximum_impact_parameter is not None else self.maximum_impact_parameter
    b = max_impact * _np.sqrt(_uniform.rvs(size=size)) # Impact parameter of the star.

    m = self.IMF.random_mass(size=size) # Mass of the star.

    zeros = _np.zeros(size)
    inc = (2.0*_np.pi * _uniform.rvs(size=size) - _np.pi) << self.units['angle'] if include_orientation else zeros
    ω = (2.0*_np.pi * _uniform.rvs(size=size) - _np.pi) << self.units['angle'] if include_orientation else zeros
    Ω = (2.0*_np.pi * _uniform.rvs(size=size) - _np.pi) << self.units['angle'] if include_orientation else zeros

    if isinstance(size, tuple): return _Stars(m=m, b=b, v=v, inc=inc, omega=ω, Omega=Ω, UNIT_SYSTEM=self.UNIT_SYSTEM, environment=self)
    elif size > 1: return _Stars(m=m, b=b, v=v, inc=inc, omega=ω, Omega=Ω, UNIT_SYSTEM=self.UNIT_SYSTEM, environment=self)
    else: return _Star(m, b[0], v[0], inc[0], ω[0], Ω[0], UNIT_SYSTEM=self.UNIT_SYSTEM)

  def random_star(self, size=1, **kwargs):
    # Alias for `random_stars`
    return self.random_stars(size=size, **kwargs)

  def stats(self):
    '''
    Prints a summary of the current stats of the Stellar Environment.
    '''
    s = self.name
    s += "\n------------------------------------------\n"
    s += "{1} Density:     {0:12.4g} \n".format(self.density, "Stellar" if self.object_unit.to_string() == _u.stars.to_string() else "Object")
    s += "Velocity Scale:      {0:12.4g} \n".format(self.velocity_dispersion)
    s += "Mass Range:            {0:6.4g} - {1:1.4g}\n".format(self.lower_mass_limit.value, self.upper_mass_limit)
    s += "Median Mass:         {0:12.4g} \n".format(self.median_mass)
    s += "Max Impact Param:    {0:12.4g} \n".format(self.maximum_impact_parameter)
    s += "Encounter Rate:      {0:12.4g} \n".format(self.encounter_rate)
    s += "------------------------------------------"
    print(s)

  def copy(self):
    '''
    Returns a deep copy of the current Stellar Environment.
    '''
    kwargs = {'stellar_density':self.density, 'velocity_dispersion':self.velocity_dispersion, 'lower_mass_limit':self.lower_mass_limit, 'upper_mass_limit':self.upper_mass_limit, 'mass_function':self.IMF.initial_mass_function, 'maximum_impact_parameter':self.maximum_impact_parameter, 'name':self.name, 'UNIT_SYSTEM':self.UNIT_SYSTEM, 'object_name':self.object_name, 'seed':self.seed, 'number_imf_samples':self.IMF.number_samples}
    return type(self)(**kwargs)

  def __eq__(self, other):
    # Overrides the default implementation
    if isinstance(other, StellarEnvironment):
        return (self.density == other.density and self.velocity_dispersion == other.velocity_dispersion and self.lower_mass_limit == other.lower_mass_limit and self.upper_mass_limit == other.upper_mass_limit and self.IMF == other.IMF and self.maximum_impact_parameter == other.maximum_impact_parameter and self.name == self.name and self.units == other.units and self.object_name == other.object_name and self.seed == other.seed)
    else:
      return NotImplemented

  def __hash__(self):
    # Overrides the default implementation
    data = []
    for d in sorted(self.__dict__.items()):
        try: data.append((d[0], tuple(d[1])))
        except: data.append(d)
    data = tuple(data)
    return hash(data)

  def summary(self, returned=False):
    '''
    Prints a compact summary of the current stats of the Stellar Environment object.
    '''
    s = f"<{self.__module__}.{type(self).__name__} object at {hex(id(self))}"
    s += f", n= {self.density:1.4g}"
    s += f", v= {self.velocity_dispersion:,.1f}"
    s += f", m= {self.lower_mass_limit.value:,.2f}-{self.upper_mass_limit.value:,.1f} {self.units['mass']}"
    s += ">"
    if returned: return s
    else: print(s)

  def __str__(self):
    return self.summary(returned=True)

  def __repr__(self):
    return self.summary(returned=True)


  @property
  def object_unit(self):
    '''The unit of the object (star) in the environment.'''
    return self.units['object']

  @property
  def object_name(self):
    '''
    Args:
      value (str): The name of the object (star) in the environment.
    '''
    return self.units['object'].to_string()

  @object_name.setter
  def object_name(self, value):
    self.units.object = _u.def_unit(value, _u.stars)

  @property
  def UNIT_SYSTEM(self):
      '''
      Args:
        value (list of Units): A list of the units to use for the environment.
      '''
      return self.units.UNIT_SYSTEM

  @UNIT_SYSTEM.setter
  def UNIT_SYSTEM(self, UNIT_SYSTEM):
    self.units.UNIT_SYSTEM = UNIT_SYSTEM

  @property
  def median_mass(self):
    '''
      The median mass of the environment's initial mass function (IMF).
    '''
    return self.IMF.median_mass.to(self.units['mass'])

  @property
  def maximum_impact_parameter(self):
    '''
      The largest impact parameter to affect a stellar system in the environment. See the examples in [Adiabatic Tests](../examples/adiabatic-tests.ipynb/#stellarenvironmentmaximum_impact_parameter) for more details.
    '''
    return self._maximum_impact_parameter.to(self.units['length'])

  @maximum_impact_parameter.setter
  def maximum_impact_parameter(self, value):
    if value is not None:
      self._maximum_impact_parameter = value.to(self.units['length']) if _tools.isQuantity(value) else value * self.units['length']
    else:
      # TODO: Convert from fminbound to interpolation
      sim = _rebound.Simulation()
      sim.add(m=1.0)
      sim.add(m=5.2e-05, a=30.2, e=0.013) # Use Neptune as a test planet.
      _f = lambda b: _np.log10(_np.abs(1e-16 - _np.abs(_analytic.relative_energy_change(sim, _Star(self.upper_mass_limit, b * self.units['length'], _np.sqrt(2.0)*_tools.maxwell_boltzmann_mean_from_dispersion(self.velocity_dispersion))))))
      bs = _np.logspace(1, 6, 1000)
      b0 = bs[_np.argmin(_f(bs))]
      self._maximum_impact_parameter = _fminbound(_f, b0/5, 5*b0) * self.units['length']

  @property
  def density(self):
    '''
      Args:
        value (Quantity): The number density of the environment. Default units: $\\rm{pc}^{-3}$.
    '''
    return self._density.to(self.units['density'])

  @density.setter
  def density(self, value):
    if _tools.isQuantity(value):
      if value.unit.is_equivalent(_u.stars/_u.m**3): self._density = value.to(self.units['density'])
      elif value.unit.is_equivalent(1/_u.m**3): self._density = (value * self.units['object']).to(self.units['density'])
      else: raise AssertionError('The given density units are not compatible.')
    else: self._density = value * self.units['density']

  @property
  def velocity_dispersion(self):
    '''
      Args:
        value (Quantity): the velocity dispersion of the environment. Default units: km/s.
    '''
    return self._velocity.to(self.units['velocity'])

  @velocity_dispersion.setter
  def velocity_dispersion(self, value):
    self._velocity = value.to(self.units['velocity']) if _tools.isQuantity(value) else value * self.units['velocity']

  @property
  def velocity_mean(self):
    '''The mean velocity of the environment. Default units: km/s.'''
    return _tools.maxwell_boltzmann_mean_from_dispersion(self.velocity_dispersion).to(self.units['velocity'])

  @property
  def velocity_mode(self):
    '''Return the most common velocity of the environment. Default units: km/s.'''
    return _tools.maxwell_boltzmann_mode_from_dispersion(self.velocity_dispersion).to(self.units['velocity'])

  @property
  def velocity_rms(self):
    '''Return the root-mean-square velocity of the environment. Default units: km/s.'''
    v = _maxwell.rvs(scale=_tools.maxwell_boltzmann_scale_from_dispersion(self.velocity_dispersion), size=int(1e6))
    return _tools.verify_unit(_np.sqrt(_np.mean(v**2)), self.units['velocity'])

  @property
  def lower_mass_limit(self):
    '''
    Args:
      value (Quantity): The lower mass limit of the initial mass function (IMF) of the environment. Default units: $M_\\odot$.
    '''
    return self.IMF.min_mass.to(self.units['mass'])

  @lower_mass_limit.setter
  def lower_mass_limit(self, value):
    self.IMF.min_mass = value

  @property
  def upper_mass_limit(self):
    '''
    Args:
      value (Quantity): The upper mass limit of the initial mass function (IMF) of the environment. Default units: $M_\\odot$.
    '''
    return self.IMF.max_mass.to(self.units['mass'])

  @upper_mass_limit.setter
  def upper_mass_limit(self, value):
    self.IMF.max_mass = value

  @property
  def IMF(self):
    '''
      Args:
        value (IMF): The initial mass function (IMF) of the environment. An `airball.IMF` object.
    '''
    return self._IMF

  @IMF.setter
  def IMF(self, value):
    if isinstance(value, _IMF): self._IMF = _IMF(value.min_mass, value.max_mass, value.imf, value.unit, value.number_samples, value.seed)
    else: raise AssertionError('Initial Mass Function (IMF) must be an airball.IMF object.')

  @property
  def encounter_rate(self):
    '''
        Compute the expected flyby encounter rate $\\Gamma = ⟨nσv⟩$ for the stellar environment in units of flybys per year.
        The inverse of the encounter rate will give the average number of years until a flyby.

        The encounter rate is computed using the following parameters:

        - n : stellar number density. Default units: $\\rm{pc}^{-3}$
        - σ : interaction cross section. Default units: $\\rm{AU}^2$
        - v : velocity dispersion. Default units: $\\rm{km/s}$

        The interaction cross section $σ = πb^2$ considers gravitational focussing where $b = q \\sqrt(1 + \\frac{2GM}{q v_∞^2})$ determined by the median mass of the environment, the maximum impact parameter, and the relative velocity at infinity derived from the velocity dispersion.
    '''
    return _tools.encounter_rate(self._density, self.velocity_mean, self._maximum_impact_parameter, self.median_mass, unit_set=self.units).to(self.units['object']/self.units['time'])

  def cumulative_encounter_times(self, size):
    '''
    Returns the cumulative time from t=0 for when to expect the next flyby encounters.
    This function assumes a Poisson Process and uses an Exponential distribution with the encounter rate.

    Args:
        size (int or tuple): The shape of the returned array. If size is an integer, it is treated as the length of the array. If size is a tuple, it is treated as the shape of the array.

    Returns:
        times (Quantity): An array of cumulative encounter times. The shape of the array is determined by the size parameter.

    Example:
        ```python
        import airball
        my_env = airball.StellarEnvironment(stellar_density=10, velocity_dispersion=20, lower_mass_limit=0.08, upper_mass_limit=8, name='My Environment')
        my_env.cumulative_encounter_times(10) # returns an array of 10 cumulative encounter times.
        ```
    '''
    if isinstance(size, tuple):
      size = tuple([int(i) for i in size])
      result = _np.cumsum(_exponential.rvs(scale=1/self.encounter_rate, size=size), axis=1) << self.units['time']
      result -= result[:, 0][:, None]
      return result
    else:
      size = int(size)
      result = _np.cumsum(_exponential.rvs(scale=1/self.encounter_rate, size=size)) << self.units['time']
      result -= result[0]
      return result

  def encounter_times(self, size):
    '''
    Returns the time between encounters for when to the expect the next flyby encounters.
    Assumes a Poisson Process and uses an Exponential distribution with the encounter rate.

    Args:
        size (int or tuple): The shape of the returned array. If size is an integer, it is treated as the length of the array. If size is a tuple, it is treated as the shape of the array.

    Returns:
        times (Quantity): An array of encounter times. The shape of the array is determined by the size parameter.

    Example:
        ```python
        import airball
        my_env = airball.StellarEnvironment(stellar_density=10, velocity_dispersion=20, lower_mass_limit=0.08, upper_mass_limit=8, name='My Environment')
        my_env.encounter_times(10) # returns an array of 10 encounter times.
        ```
    '''
    if isinstance(size, tuple):
      size = tuple([int(i) for i in size])
      return _exponential.rvs(scale=1/self.encounter_rate, size=size) << self.units['time']
    else:
      size = int(size)
      return _exponential.rvs(scale=1/self.encounter_rate, size=size) << self.units['time']

  def time_to_next_encounter(self):
    '''
    Draw a time to the next expected flyby encounter.
    Assumes a Poisson Process and uses an Exponential distribution with the encounter rate.

    Returns:
        times (Quantity): The next encounter time.

    Example:
        ```python
        import airball
        my_env = airball.StellarEnvironment(stellar_density=10, velocity_dispersion=20, lower_mass_limit=0.08, upper_mass_limit=8, name='My Environment')
        my_env.time_to_next_encounter()
        ```
    '''
    return _exponential.rvs(scale=1/self.encounter_rate) * self.units['time']



class LocalNeighborhood(StellarEnvironment):
  '''
    This is a `StellarEnvironment` subclass for the Local Neighborhood.
    It encapsulates the relevant data for a static stellar environment representing the local neighborhood of the solar system.

    The stellar density is 0.14 $\\rm{pc}^{-3}$ defined by [Bovy (2017)](https://ui.adsabs.harvard.edu/abs/2017MNRAS.470.1360B/abstract).
    The velocity distribution is defined using a Maxwell-Boltzmann distribution where the velocity dispersion is 20 km/s, defined by [Binnery & Tremaine (2008)](https://ui.adsabs.harvard.edu/abs/2008gady.book.....B/abstract) where the $v_\\rm{rms} \\sim 50$ km/s and [Bailer-Jones et al. (2018)](https://ui.adsabs.harvard.edu/abs/2018A%26A...616A..37B/abstract) so that 90% of stars have v < 100 km/s with an encounter rate of ~20 stars/Myr within 1 pc. However, a more accurate representation of the velocity distribution in the solar neighborhood is a triaxial Gaussian distribution, but that has not been implemented here.
    The mass limits is defined to between 0.08-8 solar masses using Equation (17) from [Chabrier (2003)](https://ui.adsabs.harvard.edu/abs/2003PASP..115..763C/abstract) for single stars when m < 1 and a power-law model from [Bovy (2017)](https://ui.adsabs.harvard.edu/abs/2017MNRAS.470.1360B/abstract) for stars m ≥ 1 to account for depleted stars due to stellar evolution.

    Example:
      ```python
      import airball
      my_local = airball.LocalNeighborhood()
      my_10stars = my_local.random_stars(size=10) # returns a Stars object with the masses, impact parameters, velocities, and orientation of the 10 Star objects in a heliocentric model.
      ```
  '''
  short_name = 'Local'
  def local_mass_function(x):
    '''
      This defined using Equation (17) from [Chabrier (2003)](https://ui.adsabs.harvard.edu/abs/2003PASP..115..763C/abstract) for single stars when $m < 1$ and a power-law model from [Bovy (2017)](https://ui.adsabs.harvard.edu/abs/2017MNRAS.470.1360B/abstract) for stars $m \\ge 1$ to account for depleted stars due to stellar evolution.
    '''
    chabrier03 = _imf.chabrier_2003_single(1) #0.0567
    power_law = _imf.power_law(-4.7, chabrier03(1))
    return _np.where(x > 1, power_law(x), chabrier03(x))

  def __init__(self, stellar_density = 0.14 * _u.stars/_u.pc**3, velocity_dispersion = 20.8 * _u.km/_u.s, lower_mass_limit=0.08 * _u.solMass, upper_mass_limit = 8 * _u.solMass, mass_function=local_mass_function, maximum_impact_parameter=10000 * _u.au, UNIT_SYSTEM=[], units=None, name = 'Local Neighborhood', object_name=None, seed=None, number_imf_samples=100):
    super().__init__(stellar_density=stellar_density, velocity_dispersion=velocity_dispersion, lower_mass_limit=lower_mass_limit, upper_mass_limit=upper_mass_limit, mass_function=mass_function, maximum_impact_parameter=maximum_impact_parameter, UNIT_SYSTEM=UNIT_SYSTEM, units=units, name=name, object_name=object_name, seed=seed,number_imf_samples=number_imf_samples)

class OpenCluster(StellarEnvironment):
  '''
    This is a StellarEnvironment subclass for a generic Open Cluster.
    It encapsulates the relevant data for a static stellar environment representing a generic open cluster.

    The stellar density is 100 $\\rm{pc}^{-3}$ informed by [Adams (2010)](https://ui.adsabs.harvard.edu/abs/2010ARA%26A..48...47A/abstract).
    The velocity scale is 1 km/s informed by [Adams (2010)](https://ui.adsabs.harvard.edu/abs/2010ARA%26A..48...47A/abstract) and [Malmberg, Davies, & Heggie (2011)](https://ui.adsabs.harvard.edu/abs/2011MNRAS.411..859M/abstract).
    The mass limit is defined to between 0.08-100 solar masses using Equation (17) from [Chabrier (2003)](https://ui.adsabs.harvard.edu/abs/2003PASP..115..763C/abstract) for single stars when m < 1 and [Salpeter (1955)](https://ui.adsabs.harvard.edu/abs/1955ApJ...121..161S/abstract) for stars m ≥ 1.

    Example:
      ```python
      import airball
      my_open = airball.OpenCluster()
      my_10stars = my_open.random_stars(size=10) # returns a Stars object with the masses, impact parameters, velocities, and orientation of the 10 Star objects in a heliocentric model.
      ```
  '''
  short_name = 'Open'

  def __init__(self, stellar_density = 100 * _u.stars * _u.pc**-3, velocity_dispersion = 1 * _u.km/_u.s, lower_mass_limit=0.08 * _u.solMass, upper_mass_limit = 100 * _u.solMass, mass_function=None, maximum_impact_parameter=1000 * _u.au, UNIT_SYSTEM=[], units=None, name = 'Open Cluster', object_name=None, seed=None, number_imf_samples=100):
    super().__init__(stellar_density=stellar_density, velocity_dispersion=velocity_dispersion, lower_mass_limit=lower_mass_limit, upper_mass_limit=upper_mass_limit, mass_function=mass_function, maximum_impact_parameter=maximum_impact_parameter, UNIT_SYSTEM=UNIT_SYSTEM, units=units, name=name, object_name=object_name, seed=seed,number_imf_samples=number_imf_samples)

class GlobularCluster(StellarEnvironment):
  '''
    This is a StellarEnvironment subclass for a generic Globular Cluster.
    It encapsulates the relevant data for a static stellar environment representing a generic globular cluster.

    The stellar density is 1000 $\\rm{pc}^{-3}$.
    The velocity scale is 10 km/s.
    The mass limit is defined to between 0.08-1 solar masses using Equation (17) from [Chabrier (2003)](https://ui.adsabs.harvard.edu/abs/2003PASP..115..763C/abstract) for single stars when m < 1. It is assumed that there are no stellar masses greater than 1 solar mass in a globular cluster due to stellar evolution.

    Example:
      ```python
      import airball
      my_glob = airball.GlobularCluster()
      my_10stars = my_glob.random_stars(size=10) # returns a Stars object with the masses, impact parameters, velocities, and orientation of the 10 Star objects in a heliocentric model.
      ```
  '''
  short_name = 'Globular'

  def __init__(self, stellar_density = 1000 * _u.stars * _u.pc**-3, velocity_dispersion = 10 * _u.km/_u.s, lower_mass_limit=0.08 * _u.solMass, upper_mass_limit = 1 * _u.solMass, mass_function=None, maximum_impact_parameter=5000 * _u.au, UNIT_SYSTEM=[], units=None, name = 'Globular Cluster', object_name=None, seed=None, number_imf_samples=100):
    super().__init__(stellar_density=stellar_density, velocity_dispersion=velocity_dispersion, lower_mass_limit=lower_mass_limit, upper_mass_limit=upper_mass_limit, mass_function=mass_function, maximum_impact_parameter=maximum_impact_parameter, UNIT_SYSTEM=UNIT_SYSTEM, units=units, name=name, object_name=object_name, seed=seed,number_imf_samples=number_imf_samples)

class GalacticBulge(StellarEnvironment):
  '''
    This is a StellarEnvironment subclass for a generic Galactic Bulge.
    It encapsulates the relevant data for a static stellar environment representing a generic galactic bulge. This region of the galaxy is more dense than the typical field stars found in spiral arms and has a higher velocity dispersion.

    The stellar density is 50 $\\rm{pc}^{-3}$.
    The velocity scale is 120 km/s.
    The mass limit is defined to between 0.08-10 solar masses using Equation (17) from [Chabrier (2003)](https://ui.adsabs.harvard.edu/abs/2003PASP..115..763C/abstract) for single stars when m < 1 and [Salpeter (1955)](https://ui.adsabs.harvard.edu/abs/1955ApJ...121..161S/abstract) for stars m ≥ 1.

    Example:
      ```python
      import airball
      my_bulge = airball.GalacticBulge()
      my_10stars = my_bulge.random_stars(size=10) # returns a Stars object with the masses, impact parameters, velocities, and orientation of the 10 Star objects in a heliocentric model.
      ```
  '''
  short_name = 'Bulge'

  def __init__(self, stellar_density = 50 * _u.stars * _u.pc**-3, velocity_dispersion = 120 * _u.km/_u.s, lower_mass_limit=0.08 * _u.solMass, upper_mass_limit = 10 * _u.solMass, mass_function=None, maximum_impact_parameter=50000 * _u.au, UNIT_SYSTEM=[], units=None, name = 'Milky Way Bulge', object_name=None, seed=None, number_imf_samples=100):
    super().__init__(stellar_density=stellar_density, velocity_dispersion=velocity_dispersion, lower_mass_limit=lower_mass_limit, upper_mass_limit=upper_mass_limit, mass_function=mass_function, maximum_impact_parameter=maximum_impact_parameter, UNIT_SYSTEM=UNIT_SYSTEM, units=units, name=name, object_name=object_name, seed=seed,number_imf_samples=number_imf_samples)

class GalacticCore(StellarEnvironment):
  '''
    This is a StellarEnvironment subclass for a generic Galactic Core.
    It encapsulates the relevant data for a static stellar environment representing a generic galactic core. This is the densest region of the galaxy and has the highest velocity dispersion.

    The stellar density is $10^4$ $\\rm{pc}^{-3}$.
    The velocity scale is 170 km/s.
    The mass limit is defined to between 0.08-10 solar masses using Equation (17) from [Chabrier (2003)](https://ui.adsabs.harvard.edu/abs/2003PASP..115..763C/abstract) for single stars when m < 1 and [Salpeter (1955)](https://ui.adsabs.harvard.edu/abs/1955ApJ...121..161S/abstract) for stars m ≥ 1.

    Example:
      ```python
      import airball
      my_core = airball.GalacticCore()
      my_10stars = my_core.random_stars(size=10) # returns a Stars object with the masses, impact parameters, velocities, and orientation of the 10 Star objects in a heliocentric model.
      ```
  '''
  short_name = 'Core'

  def __init__(self, stellar_density = 10000 * _u.stars * _u.pc**-3, velocity_dispersion = 170 * _u.km/_u.s, lower_mass_limit=0.08 * _u.solMass, upper_mass_limit = 10 * _u.solMass, mass_function=None, maximum_impact_parameter=50000 * _u.au, UNIT_SYSTEM=[_u.yr], units=None, name = 'Milky Way Core', object_name=None, seed=None, number_imf_samples=100):
    super().__init__(stellar_density=stellar_density, velocity_dispersion=velocity_dispersion, lower_mass_limit=lower_mass_limit, upper_mass_limit=upper_mass_limit, mass_function=mass_function, maximum_impact_parameter=maximum_impact_parameter, UNIT_SYSTEM=UNIT_SYSTEM, units=units, name=name, object_name=object_name, seed=seed,number_imf_samples=number_imf_samples)



import numpy as _np
import types as _types
from scipy.integrate import quad as _quad
from . import units as _u
from . import tools as _tools

class Distribution():
  '''
  Base class for defining a probability distribution function.
  This class is used to wrap a probability distribution function and its parameters into a single object.

  Args:
    mass_function (function): Probability distribution function.
    args (list): List of arguments for the probability distribution function.

  Returns:
    pdf (float or ndarray): Probability density at the given mass value(s).
  
  Example:
    ```python
    import airball
    mf = airball.imf.Distribution(lambda x, A: A * x, [1])
    imf = airball.IMF(0.1, 100, mass_function=mf)
    imf.random_mass()
    ```

  '''

  def __init__(self, mass_function, args):
    self.mass_function = mass_function
    self.params = args
  
  def __call__(self, x):
    return self.mass_function(x, *self.params)

class chabrier_2003_single(Distribution):
  """
  Chabrier 2003 IMF for single stars.
  This function calculates the probability density for a given mass value (x) based on Equation (17) of Chabrier 2003 for an IMF for single stars.

  $$PDF(x) = \\frac{A}{x} \\exp\\left[-\\frac{(\\log_{10}(x) - \\log_{10}(0.079))^2 }{ 2 \\cdot 0.69^2}\\right]$$

  Args:
    A (float, optional): Normalization factor. Default is 0.158.

  Returns:
    pdf (float or ndarray): Probability density at the given mass value(s).

  Example:
    ```python
    import airball
    imf = airball.IMF(0.1, 100, mass_function=airball.imf.chabrier_2003_single(A=0.158))
    imf.random_mass()
    ```
  """
  def __init__(self, A=0.158):
    super().__init__(self._chabrier_2003_single, [A])
  
  def _chabrier_2003_single(self, x, A=0.158):
    return (A / x) * _np.exp(-((_np.log10(x) - _np.log10(0.079)) ** 2) / (2 * 0.69 ** 2))

class salpeter_1955(Distribution):
  """
  Salpeter 1955 IMF for single stars.
  This function calculates the probability density for a given mass value (x) based on the Salpeter 1955 IMF equation.

  $$PDF(x) = A x^{-2.3}$$

  Args:
    A (float): Normalization factor.

  Returns:
    pdf (float or ndarray): Probability density at the given mass value(s).
  
  Example:
    ```python
    import airball
    imf = airball.IMF(0.1, 100, mass_function=airball.imf.salpeter_1955(A=1))
    imf.random_mass()
    ```
  """
  def __init__(self, A):
    super().__init__(self._salpeter_1955, [A])
  
  def _salpeter_1955(self, x, A):
    return A * x ** -2.3

class default_mass_function(Distribution):
  '''
  Default mass function for an IMF. This function is a piecewise function that uses the IMF for single stars from Chabrier 2003 for masses less than 1 solar mass and the Salpeter 1955 IMF for masses greater than 1 solar mass.

  Args:
    x (float or ndarray): Mass value(s).

  Returns:
    pdf (float or ndarray): Probability density at the given mass value(s).

  Example:
    ```python
    import airball
    imf = airball.IMF(0.1, 100, mass_function=airball.imf.default_mass_function())
    imf.random_mass()
    ```
  '''
  def __init__(self):
    super().__init__(self._default_mass_function, [])
  
  def _default_mass_function(self, x):
    chabrier03 = chabrier_2003_single(A=0.158)
    salpeter55 = salpeter_1955(A=chabrier03(1))
    return _np.where(x < 1, chabrier03(x), salpeter55(x))

class uniform(Distribution):
  '''
  Uniform IMF.
  This function calculates the probability density for a given mass value (x) based on a uniform IMF.
  
  $$PDF(x) = 1$$
  '''
  def __init__(self):
    super().__init__(self._uniform)
  
  def _uniform(self, x):
    return x * 0 + 1

class power_law(Distribution):
  '''
  Power law IMF.
  This function calculates the probability density for a given mass value (x) based on a power law IMF.
  
  $$PDF(x) = A x^\\alpha$$
  
  Args:
    alpha (float): Power law index.
    A (float): Normalization factor.

  Returns:
    pdf (float or ndarray): Probability density at the given mass value(s).
  '''
  def __init__(self, alpha, A):
    super().__init__(self._power_law, [alpha, A])
  
  def _power_law(self, x, alpha, A):
    return A * x ** alpha

class broken_power_law(Distribution):
  '''
  Broken power law IMF.
  This function calculates the probability density for a given mass value (x) based on a broken power law IMF.
  
  $$PDF(x) = \\begin{cases} A x^\\alpha & x < x_0 \\\\ A x_0^{\\beta - \\alpha} x^\\beta & x \\geq x_0 \\end{cases}$$
  
  Args:
    alpha (float): Power law index for $x < x_0$.
    beta (float): Power law index for $x ≥ x_0$.
    A (float): Normalization factor.
    x_0 (float): Break point between the two power laws.
    
  Returns:
    pdf (float or ndarray): Probability density at the given mass value(s).
  '''
  def __init__(self, alpha, beta, A, x_0):
    super().__init__(self._broken_power_law, [alpha, beta, A, x_0])

  def _broken_power_law(self, x, alpha, beta, A, x_0):
    return _np.where(x < x_0, A * x ** alpha, A * x_0 ** (beta - alpha) * x ** beta)

class lognormal(Distribution):
  '''
  Lognormal IMF.
  This function calculates the probability density for a given mass value (x) based on a lognormal IMF.
  
  $$PDF(x) = A \\exp\\left[-\\frac{(\\log_{10}(x) - \\mu)^2 }{ 2 \\sigma^2}\\right]$$
  
  Args:
    mu (float): Mean of the lognormal distribution.
    sigma (float): Standard deviation of the lognormal distribution.
    A (float): Normalization factor.
    
  Returns:
    pdf (float or ndarray): Probability density at the given mass value(s).
  '''
  def __init__(self, mu, sigma, A):
    super().__init__(self._lognormal, [mu, sigma, A])
  
  def _lognormal(self, x, mu, sigma, A):
    return A * _np.exp(-(x - mu) ** 2 / (2 * sigma ** 2))

class loguniform(Distribution):
  '''
  Loguniform IMF.
  This function calculates the probability density for a given mass value (x) based on a loguniform IMF.
  
  $$PDF(x) = \\frac{A}{x}$$
  
  Args:
    A (float): Normalization factor.
    
  Returns:
    pdf (float or ndarray): Probability density at the given mass value(s).
  '''
  def __init__(self, A):
    super().__init__(self._loguniform, [A])

  def _loguniform(self, x, A):
    return A / x

class IMF():
  """
  Class representing an Initial Mass Function (IMF).
  It generates random masses based on a given mass function and provides various properties and methods for manipulating and analyzing the IMF.

  Args:
    min_mass (float): Minimum mass value of the IMF range.
    max_mass (float): Maximum mass value of the IMF range.
    mass_function (function, optional): Mass function to use for the IMF. Default is a piecewise Chabrier 2003 and Salpeter 1955.
    unit (Unit, optional): Unit of mass. Default is solar masses.
    number_samples (float, optional): Number of samples to use for interpolating the CDF. Default is 100.
    seed (float, optional): Value to seed the random number generator with. Default is None.

  Attributes:
    min_mass (float): Minimum mass value of the IMF range.
    max_mass (float): Maximum mass value of the IMF range.
    median_mass (float): Median mass value of the IMF.
    seed (float): Value to seed the random number generator with.
    number_samples (float): Number of samples to use for interpolating the CDF.
    unit (Unit): Unit of mass.
    normalization_factor (float): Normalization factor for the PDF.
    masses (ndarray): Mass values logarithmically spanning the IMF range.
    CDF (function): Cumulative distribution function (CDF) of the IMF.
    PDF (function): Normalized probability density function (PDF) of the IMF.
    IMF (function): Initial mass function (IMF) of the IMF.
  """

  def __init__(self, min_mass, max_mass, mass_function=None, unit=_u.solMass, number_samples=100, seed=None):
    self._number_samples = int(number_samples)
    self._seed = seed
    self.unit = unit if _u.isUnit(unit) else _u.solMass

    # Convert min_mass and max_mass to specified unit if they are Quantity objects, otherwise assume they are already in the correct unit
    self._min_mass = min_mass.to(self.unit) if _tools.isQuantity(min_mass) else min_mass * self.unit
    if self._min_mass.value <= 0: raise ValueError('Cannot have minimum mass value be less than or equal to 0.')
    self._max_mass = max_mass.to(self.unit) if _tools.isQuantity(max_mass) else max_mass * self.unit
    if self._max_mass <= self._min_mass: raise ValueError('Cannot have maximum mass value be less than or equal to minimum mass.')
    if self._max_mass.value <= 0: raise ValueError('Cannot have maximum mass value be less than or equal to 0.')

    # Determine the probability distribution function (PDF) based on the given mass function or default to a piecewise Chabrier 2003 and Salpeter 1955
    if mass_function is None: mass_function = default_mass_function()
    elif not isinstance(mass_function, (_types.FunctionType, Distribution)): raise ValueError('mass_function must be a function or a Distribution object.')
    self.initial_mass_function = mass_function

    # Recalculate the IMF properties based on the updated parameters
    self._recalculate()

  ### Define the PDF, normalized PDF, and CDF functions for the IMF. These functions are vectorized so they can accept arrays of values. ###
  def _pdf_(self, x): return _np.where(x < self.min_mass.value, 0, _np.where(x > self.max_mass.value, 0, self.imf(x)))
  def _probability_density_function(self, x): return _np.vectorize(self._pdf_)(x)
  def _npdf_(self, x): return self._pdf_(x) / self.normalization_factor
  def _normalized_probability_density_function(self, x): return _np.vectorize(self._npdf_)(x)
  def _cdf_(self, x): return _np.where(x < self.min_mass.value, 0, _np.where(x > self.max_mass.value, 1,  _quad(self._normalized_probability_density_function, self.min_mass.value, x)[0]))
  def _cumulative_distribution_function(self, x): return _np.vectorize(self._cdf_)(x)

  def _recalculate(self):
    """
    Recalculates the IMF properties based on the current parameters.
    This function updates the normalization factor, normalized PDF, cumulative distribution function (CDF),
    mass values, and IMF values based on the current min_mass, max_mass, and PDF.
    """
    # Calculate the normalization factor for the PDF
    self.normalization_factor = _quad(self._probability_density_function, self._min_mass.value, self._max_mass.value)[0]
    # Generate logarithmically spaced mass values between min_mass and max_mass
    self._masses = _np.logspace(_np.log10(self._min_mass.value), _np.log10(self._max_mass.value), self.number_samples)
    # Calculate the CDF for the mass values
    self._CDF = self._cumulative_distribution_function(self._masses)

  def random_mass(self, size=1, **kwargs):
    """
    Generates random mass values from the IMF.

    Args:
      size (int or tuple): Number of mass values to generate. If size is a tuple, it is interpreted as array dimensions. Default: 1.

    Keyword Args:
      seed (int): Seed for the random number generator. Default: None.

    Returns:
      masses (float or ndarray): Randomly generated mass value(s) from the IMF.

    Example:
      ```python
      import airball
      imf = airball.IMF(0.1, 100)
      imf.random_mass()
      ```
    """
    if 'seed' in kwargs: self.seed = kwargs['seed']
    if self.seed != None: _np.random.seed(self.seed)
    
    if isinstance(size, tuple): size = tuple([int(i) for i in size])
    else: size = int(size)

    rand_masses = _np.interp(_np.random.uniform(size=size), self._CDF, self._masses) * self.unit
    if isinstance(size, tuple): return rand_masses
    elif size > 1: return rand_masses
    else: return rand_masses[0]

  def masses(self, number_samples, endpoint=True):
    """
    Convenience function for generating an array of mass values logarithmically spanning the IMF range.

    Args:
      number_samples (int): Number of mass values to generate.
      endpoint (bool, optional): Whether to include the max_mass value in the array. Default: True.

    Returns:
      masses (ndarray): numpy array of mass values logarithmically spanning the IMF range.
    """
    return _np.logspace(_np.log10(self._min_mass.value), _np.log10(self._max_mass.value), int(number_samples), endpoint=endpoint)

  @property
  def min_mass(self):
    """
    The minimum mass value of the IMF range.
    Recalculates the IMF properties when the `min_mass` value is updated.

    Args:
      value (float): New minimum mass value. Must be greater than 0. Units are assumed to be the same as the IMF unit.
    """
    return self._min_mass

  @min_mass.setter
  def min_mass(self, value):
    value = value.to(self.unit) if _tools.isQuantity(value) else value * self.unit
    if value.value <= 0: raise ValueError('Cannot have minimum mass value be less than or equal to 0.')
    self._min_mass = value
    self._recalculate()

  @property
  def max_mass(self):
    """
    The maximum mass value of the IMF range.
    Recalculates the IMF properties when the `max_mass` value is updated.

    Args:
      value (float): New maximum mass value. Must be greater than minimum mass value. Units are assumed to be the same as the IMF unit.
    """
    return self._max_mass

  @max_mass.setter
  def max_mass(self, value):
    value = value.to(self.unit) if _tools.isQuantity(value) else value * self.unit
    if value.value <= 0: raise ValueError('Cannot have maximum mass value be less than or equal to 0.')
    if value.value <= self.min_mass: raise ValueError('Cannot have maximum mass value be less than or equal to minimum mass value.')
    self._max_mass = value.to(self.unit) if _tools.isQuantity(value) else value * self.unit
    self._recalculate()

  @property
  def median_mass(self):
    """
    Median mass value of the IMF.
    """
    return _np.interp(0.5, self._CDF, self._masses) * self.unit
  
  @property
  def seed(self):
    """
    The seed for the random number generator 

    Args:
      value (int or None): New seed for the random number generator (int, or None to turn off).
    """
    return self._seed

  @seed.setter
  def seed(self, value):
    self._seed = value

  @property
  def number_samples(self):
    """
    The number of samples to use for interpolating the CDF.
    Recalculates the IMF properties when the `number_samples` value is updated.

    Args:
      value (int): New number of samples to use for interpolating the CDF.
    """
    return self._number_samples

  @number_samples.setter
  def number_samples(self, value):
    self._number_samples = int(value)
    self._recalculate()

  def cdf(self, x):
    """Cumulative distribution function (CDF) of the IMF."""
    return _np.interp(x, self._masses, self._CDF)

  @property
  def CDF(self):
    return self.cdf

  def pdf(self, x):
    """Normalized probability density function (PDF) of the IMF."""
    return self._normalized_probability_density_function(x)

  @property
  def PDF(self):
    return self.pdf

  def imf(self, x):
    """The initial mass function (IMF) of the IMF."""
    return self.initial_mass_function(x)

  @property
  def IMF(self):
    return self.imf

  def __eq__(self, other):
    # Overrides the default implementation
    if isinstance(other, IMF):
        return (self.min_mass == other.min_mass and self.max_mass == other.max_mass and self.initial_mass_function == other.initial_mass_function and self.unit == other.unit and self.number_samples == other.number_samples and self.seed==other.seed)
    else:
      return NotImplemented
  
  def __hash__(self):
    # Overrides the default implementation
    data = []
    for d in sorted(self.__dict__.items()):
        try: data.append((d[0], tuple(d[1])))
        except: data.append(d)
    data = tuple(data)
    return hash(data)
  

  def summary(self, returned=False):
    ''' 
    Prints a compact summary of the current stats of the Stellar Environment object.
    '''
    s = f"<{self.__module__}.{type(self).__name__} object at {hex(id(self))}"
    s += f", m= {self.min_mass.value:,.2f}-{self.max_mass.value:,.1f} {self.unit}"
    s += f", IMF= {self.initial_mass_function.__class__.__name__}" if isinstance(self.initial_mass_function, Distribution) else ', IMF= custom'
    s += ">"
    if returned: return s
    else: print(s)
  
  def __str__(self):
    return self.summary(returned=True)
  
  def __repr__(self):
    return self.summary(returned=True)
